#ifndef FUNCSIG_H
#define FUNCSIG_H

#define FUNCSIG_SZ 33

typedef struct funcsig {
    uint8_t data[FUNCSIG_SZ];
    uint8_t undef[FUNCSIG_SZ];
} funcsig;

typedef struct FunctionSignature {
    char *name;
    uint64_t offset;
} FunctionSignature;

FunctionSignature *find_function_signatures(FILE *f);

static const funcsig FUNCSIGS_MALLOC[] = {
    
    // test-static2
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x54, // push r12,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x97ad3] # 0xae7d0,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x16e48,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x88, 0x00, // js 0x16e28
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff // len=33
        }
    },

    // libc-2.33-5.fc34.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x136b3d] # 0x1c2ed8,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x8c508,
            0x48, 0x85, 0xff, // test rdi, rdi
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6-amd64_2.24-3ubuntu1_i386.so
    {
        .data = {
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x08, // sub rsp, 0x8,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x31e0f3] # 0x398ef0,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x7af48,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x31df80] # 0x398d90,
            0x48, // mov rbp, rdi
        }, 
        .undef = {
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // libc6_2.13-20ubuntu5.2_amd64.so
    {
        .data = {
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x10], r12,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x20], rbx,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x18], rbp,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x8], r13,
            0x48, 0x83, 0xec, 0x28, // sub rsp, 0x28,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x31d3e6] # 0x39be98
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff // len=33
        }
    },

    // libc6_2.17-93ubuntu2_amd64.so
    {
        .data = {
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x08, // sub rsp, 0x8,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x33da60] # 0x3c0ee0,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x835ad,
            0x48, 0x8b, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x33d8ed] # 0x3c0d80
        }, 
        .undef = {
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff // len=33
        }
    },

    // musl-1.1.12-r8.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x18], rdi,
            0x48, 0x8d, 0x7c, 0x24, 0x18, // lea rdi, [rsp+0x18],
            0xe8, 0x07, 0x00, 0x00, 0x00, // call 0x266aa,
            0x85, 0xc0, // test eax, eax,
            0x00, 0x00, // js 0x2702a
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0xff, 0xff // len=33
        }
    },

    // libc6_2.7-9ubuntu2_amd64.so
    {
        .data = {
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x10], rbp,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x18], rbx,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x8], r12,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x2e5ad3] # 0x35be10,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, // test rax, rax
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00 // len=33
        }
    },

    // libc6_2.31-17_amd64.so
    {
        .data = {
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x133df9] # 0x1bdef0,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x8a250,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x88, 0x00, 0x00, // js 0x8a230
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff // len=33
        }
    },

    // libc6_2.15-0ubuntu1_amd64.so
    {
        .data = {
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x18], rbp,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x20], rbx,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x10], r12,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x8], r13,
            0x48, 0x83, 0xec, 0x28, // sub rsp, 0x28,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x3331ee] # 0x3b3ea0
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff // len=33
        }
    },

    // libc-2.26-lp150.11.9.1.x86_64.so
    {
        .data = {
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x32f503] # 0x3b3ef0,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x84c08,
            0x48, 0x8d, 0x47, 0x17, // lea rax, [rdi+0x17],
            0x48, 0x83, 0xf8, 0x1f, // cmp rax, 0x1f
        }, 
        .undef = {
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.13-0ubuntu15_amd64.so
    {
        .data = {
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x10], r12,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x20], rbx,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x18], rbp,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x8], r13,
            0x48, 0x83, 0xec, 0x48, // sub rsp, 0x48,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x322bae] # 0x39ee00
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff // len=33
        }
    },

    // musl-1.2.2-3-omv4050.x86_64.so
    {
        .data = {
            0x50, // push rax,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x89, 0x00, 0x00, // mov QWORD PTR [rsp], rax,
            0xe8, 0xc7, 0x00, 0x00, 0x00, // call 0x5f48e,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rcx, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x3b, 0x00, 0x00, // cmp rcx, QWORD PTR [rsp],
            0x00, // jne 0x5e6d8
        }, 
        .undef = {
            0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0xff // len=33
        }
    },

    // musl-1.1.22-r4.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x28, // sub rsp, 0x28,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x8], rdi,
            0x48, 0x8d, 0x7c, 0x24, 0x08, // lea rdi, [rsp+0x8],
            0xe8, 0x55, 0x00, 0x00, 0x00, // call 0x22805,
            0x85, 0xc0, // test eax, eax,
            0x79, 0x07, // jns 0x22dbb
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // libc6_2.34-0ubuntu1_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x10, // sub rsp, 0x10,
            0x80, 0x3d, 0x32, 0x7a, 0x17, 0x00, 0x00, // cmp BYTE PTR [rip+0x177a32], 0x0 # 0x2201a8,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0xa8958,
            0x48, 0x85, 0xed, // test rbp, rbp,
            0x00, 0x00, // js 0xa8966
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff // len=33
        }
    },

    // libc6_2.29-0ubuntu2_amd64.so
    {
        .data = {
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x08, // sub rsp, 0x8,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x14b49f] # 0x1e3ef0,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x98c40,
            0x48, 0x8d, 0x47, 0x17, // lea rax, [rdi+0x17]
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.3.2.ds1-13ubuntu2.2_amd64_2.so
    {
        .data = {
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rdx, QWORD PTR [rip+0x1c48c5] # 0x23b5b0,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x10], r12,
            0x48, 0x89, 0x00, 0x00, // mov QWORD PTR [rsp], rbx,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x8], rbp,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rdx],
            0x48, 0x85, // test rax, rax
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // musl-1.1.24-r3.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x28, // sub rsp, 0x28,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x8], rdi,
            0x48, 0x8d, 0x7c, 0x24, 0x08, // lea rdi, [rsp+0x8],
            0xe8, 0x47, 0x00, 0x00, 0x00, // call 0x22bbd,
            0x85, 0xc0, // test eax, eax,
            0x79, 0x08, // jns 0x23182
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // musl-1.2.2-r7.so
    {
        .data = {
            0xe9, 0xc6, 0x13, 0x00, 0x00, // jmp 0x24851,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0x29, 0xf8, // sub rax, rdi,
            0x29, 0xd0, // sub eax, edx,
            0x00, 0x00, // je 0x234b1,
            0x89, 0xc2, // mov edx, eax,
            0xf7, 0xda, // neg edx,
            0x48, 0x63, 0xd2, // movsxd rdx, edx,
            0xc6, 0x04, 0x16, 0x00, // mov BYTE PTR [rsi+rdx*1], 0x0,
            0x83, 0xf8, 0x04, // cmp eax, 0x4,
            0x00, 0x00, // jle 0x234b1,
            0xc6, 0x46, // mov BYTE PTR [rsi-0x5], 0x0
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // glibc-2.28-160.el8.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x33944d] # 0x3beee8,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x85c10,
            0x48, 0x8d, 0x47, 0x17, // lea rax, [rdi+0x17],
            0x48, 0x83, // cmp rax, 0x1f
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // musl_1.2.1-1_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0xb8, 0xfe, 0xef, 0xff, 0xff, 0xff, // movabs rax, 0x7fffffffffffeffe,
            0x00, 0x00, 0x00, // ,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x39, 0xc7, // cmp rdi, rax,
            0x0f, 0x87, 0x69, 0x03, 0x00, 0x00, // ja 0x2c0e8,
            0x48, 0x89, // mov rbp, rdi
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // libc6_2.3.2.ds1-13ubuntu2.2_amd64.so
    {
        .data = {
            0x48, 0x83, 0xec, 0x28, // sub rsp, 0x28,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1c0705] # 0x2340d0,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x18], r12,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x8], rbx,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x10], rbp,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x20], r13,
            0x49, 0x89, // mov r12, rdi
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // libc-2.34-11.mga9.x86_64_3.so
    {
        .data = {
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x4a71] # 0x9fa8,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x5670,
            0x8b, 0x1d, 0xa8, 0x55, 0x00, 0x00, // mov ebx, DWORD PTR [rip+0x55a8] # 0xaaf8,
            0xf6, // test bl, 0x1
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00 // len=33
        }
    },

    // musl-1.1.11-r5.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x18], rdi,
            0x48, 0x8d, 0x7c, 0x24, 0x18, // lea rdi, [rsp+0x18],
            0xe8, 0xe7, 0x00, 0x00, 0x00, // call 0x267c3,
            0x85, 0xc0, // test eax, eax,
            0x00, 0x00, // js 0x27166
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0xff, 0xff // len=33
        }
    },

    // libc-2.29-1-x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x08, // sub rsp, 0x8,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x13a403] # 0x1beed8,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x84c70
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff // len=33
        }
    },

    // libc-2.34-0-omv4050.x86_64.so
    {
        .data = {
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x10, // sub rsp, 0x10,
            0x80, 0x3d, 0x7e, 0xe2, 0x16, 0x00, 0x00, // cmp BYTE PTR [rip+0x16e27e], 0x0 # 0x212ff0,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0xa4ec0,
            0x48, 0x85, 0xed, // test rbp, rbp,
            0x0f, 0x88, 0x00, 0x00, 0x00, 0x00, // js 0xa4ece
        }, 
        .undef = {
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff // len=33
        }
    },

    // musl_1.1.24-1_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x18], rdi,
            0x48, 0x8d, 0x7c, 0x24, 0x18, // lea rdi, [rsp+0x18],
            0x64, 0x48, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff // len=33
        }
    },

    // libc6-amd64_2.9-4ubuntu6.3_i386.so
    {
        .data = {
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x18], rbx,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x10], rbp,
            0x48, 0x89, 0xfb, // mov rbx, rdi,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x8], r12,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x2ef61b] # 0x36bdf8,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, // test rax, rax
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00 // len=33
        }
    },

    // libc-2.29-3-x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x08, // sub rsp, 0x8,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x134908] # 0x1bbef0,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, // jne 0x87790
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff // len=33
        }
    },

    // libc6_2.26-0ubuntu2_amd64.so
    {
        .data = {
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x34b049] # 0x3d9ee0,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x8f0b8,
            0x48, 0x8d, 0x47, 0x17, // lea rax, [rdi+0x17],
            0x31, 0xdb, // xor ebx, ebx,
            0x48, 0x83, 0xf8, 0x1f, // cmp rax, 0x1f
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.32-18.mga8.x86_64.so
    {
        .data = {
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x08, // sub rsp, 0x8,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x12c98c] # 0x1b2ed0,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x86660,
            0x48, // test rdi, rdi
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // musl-1.1.22-2-x86_64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x48, 0x8d, 0x57, 0xff, // lea rdx, [rdi-0x1],
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x28], rax,
            0x31, // xor eax, eax
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // musl_0.9.15-1_amd64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x48, 0x8d, 0x57, 0xff, // lea rdx, [rdi-0x1],
            0x48, 0xb8, 0xdf, 0xef, 0xff, 0xff, 0xff, // movabs rax, 0x7fffffffffffefdf,
            0x00, 0x00, 0x00, // ,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x28, // sub rsp, 0x28,
            0x48, 0x39, 0xc2, // cmp rdx, rax,
            0x0f, 0x86, // jbe 0x24160
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // libc6-amd64_2.26-0ubuntu2.1_i386.so
    {
        .data = {
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x32d149] # 0x3aeee0,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x81fa0,
            0x48, 0x83, 0xff, 0xbf, // cmp rdi, 0xffffffffffffffbf,
            0x0f, 0x87, 0xff, 0x01, 0x00, 0x00, // ja 0x81fb0
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.34-0ubuntu2_amd64_2.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x49, 0x89, 0xfd, // mov r13, rdi,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x30, // sub rsp, 0x30,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x28], rax
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff // len=33
        }
    },

    // musl_1.1.5-2+deb8u1_amd64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x48, 0x8d, 0x57, 0xff, // lea rdx, [rdi-0x1],
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x48, 0xb8, 0xdf, 0xef, 0xff, 0xff, 0xff, // movabs rax, 0x7fffffffffffefdf,
            0x00, 0x00, 0x00, // ,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x28, // sub rsp, 0x28,
            0x48, 0x39, 0xc2, // cmp rdx, rax,
            0x0f, 0x86, // jbe 0x26fd8
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0xff, 0xff, 0xff,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // musl-1.1.24-r10.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x18], rdi,
            0x48, 0x8d, 0x7c, 0x24, 0x18, // lea rdi, [rsp+0x18],
            0xe8, 0xfe, 0x00, 0x00, 0x00, // call 0x22ba6,
            0x85, 0xc0, // test eax, eax,
            0x79, 0x08, // jns 0x231b4
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // libc6-amd64_2.34-0ubuntu3_i386_2.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x64ad] # 0xbfa8,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x5c30,
            0x8b, 0x1d, 0x04, // mov ebx, DWORD PTR [rip+0x7004] # 0xcb18
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // musl-1.1.19-r11.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x18], rdi,
            0x48, 0x8d, 0x7c, 0x24, 0x18, // lea rdi, [rsp+0x18],
            0xe8, 0xff, 0x00, 0x00, 0x00, // call 0x23e67,
            0x85, 0xc0, // test eax, eax,
            0x79, 0x07, // jns 0x24473
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // libc-2.34-7.fc35.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x10, // sub rsp, 0x10,
            0x80, 0x3d, 0xb2, 0x49, 0x16, 0x00, 0x00, // cmp BYTE PTR [rip+0x1649b2], 0x0 # 0x201fe8,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x9d810,
            0x48, 0x85, 0xed, // test rbp, rbp,
            0x00, 0x00, // js 0x9d81e
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff // len=33
        }
    },

    // musl_1.1.21-2_amd64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x18], rdi,
            0x48, 0x8d, 0x7c, 0x24, 0x18, // lea rdi, [rsp+0x18],
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // 
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff // len=33
        }
    },

    // libc-2.34-11.mga9.x86_64.so
    {
        .data = {
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x80, 0x3d, 0x64, 0x7b, 0x15, 0x00, 0x00, // cmp BYTE PTR [rip+0x157b64], 0x0 # 0x1ee1a8,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x96790,
            0x48, 0x85, 0xed, // test rbp, rbp,
            0x0f, 0x88, 0x00, 0x00, // js 0x9679e
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff // len=33
        }
    },

    // libc-2.34-7.fc35.x86_64_2.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x646d] # 0xbfa8,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x5c70,
            0x8b, 0x1d, 0x64, // mov ebx, DWORD PTR [rip+0x6e64] # 0xc9b8
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.34-0ubuntu3_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x10, // sub rsp, 0x10,
            0x80, 0x3d, 0x02, 0x78, 0x17, 0x00, 0x00, // cmp BYTE PTR [rip+0x177802], 0x0 # 0x2201a8,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0xa8b88,
            0x48, 0x85, 0xed, // test rbp, rbp,
            0x00, 0x00, // js 0xa8b96
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff // len=33
        }
    },

    // libc-2.31-7.30.x86_64.so
    {
        .data = {
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x343fd9] # 0x3cdef0,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x8a060,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x88, 0x00, 0x00, 0x00, 0x00, // js 0x8a070,
            0x48, // lea rax, [rdi+0x17]
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // musl-1.2.2-1-x86_64.so
    {
        .data = {
            0xe9, 0xbe, 0x05, 0x01, 0x00, // jmp 0x27cb3,
            0x66, 0x2e, 0x0f, 0x1f, 0x84, 0x00, 0x00, // cs nop WORD PTR [rax+rax*1+0x0],
            0x00, 0x00, 0x00, // ,
            0x90, // nop,
            0xe9, 0x4b, 0xd9, 0xff, 0xff, // jmp 0x15050,
            0x66, 0x2e, 0x0f, 0x1f, 0x84, 0x00, 0x00, // cs nop WORD PTR [rax+rax*1+0x0],
            0x00, 0x00, 0x00, // ,
            0x90, // nop,
            0x53, // push rbx
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0xff, 0xff, 0xff,
            0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0xff, 0xff, 0xff,
            0x00,
            0x00 // len=33
        }
    },

    // musl-1.1.5-r5.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x28, // sub rsp, 0x28,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x18], rdi,
            0x48, 0x8d, 0x7c, 0x24, 0x18, // lea rdi, [rsp+0x18],
            0xe8, 0xed, 0x00, 0x00, 0x00, // call 0x25aee,
            0x85, 0xc0, // test eax, eax,
            0x79, 0x07, // jns 0x2630c
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // libc-2.34-1.2.x86_64_2.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x647d] # 0xbfa8,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x5c60,
            0x8b, 0x1d, 0x94, // mov ebx, DWORD PTR [rip+0x6e94] # 0xc9d8
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // musl-1.2.1-2-x86_64.so
    {
        .data = {
            0x48, 0xb8, 0xfe, 0xef, 0xff, 0xff, 0xff, // movabs rax, 0x7fffffffffffeffe,
            0x00, 0x00, 0x00, // ,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x39, 0xc7, // cmp rdi, rax,
            0x77, 0x11, // ja 0x27a2a,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x48, 0x81, 0xff, 0xeb, 0xff, 0x01, 0x00, // cmp rdi, 0x1ffeb
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.26-0ubuntu2.1_amd64.so
    {
        .data = {
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x34b029] # 0x3d9ee0,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x8f0e8,
            0x48, 0x83, 0xff, 0xbf, // cmp rdi, 0xffffffffffffffbf,
            0x0f, 0x87, 0x27, 0x02, 0x00, 0x00, // ja 0x8f0f8
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6-amd64_2.34-0ubuntu3_i386.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x10, // sub rsp, 0x10,
            0x80, 0x3d, 0xa2, 0x4a, 0x16, 0x00, 0x00, // cmp BYTE PTR [rip+0x164aa2], 0x0 # 0x2021a8,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x9d8e8,
            0x48, 0x85, 0xed, // test rbp, rbp,
            0x00, 0x00, // js 0x9d8f6
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff // len=33
        }
    },

    // musl-1.2.2-1-omv4002.x86_64.so
    {
        .data = {
            0x50, // push rax,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x89, 0x00, 0x00, // mov QWORD PTR [rsp], rax,
            0xe8, 0xd0, 0x00, 0x00, 0x00, // call 0x5f42c,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rcx, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x3b, 0x00, 0x00, // cmp rcx, QWORD PTR [rsp],
            0x00, // jne 0x5e66d
        }, 
        .undef = {
            0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0xff // len=33
        }
    },

    // libc-2.34-0-omv4050.x86_64_2.so
    {
        .data = {
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x49, 0x89, 0xfd, // mov r13, rdi,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x30, // sub rsp, 0x30,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x28], rax,
            0x48, 0x8b, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x5a34] # 0xafa8
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff // len=33
        }
    },

    // libc6_2.26-0ubuntu4_amd64.so
    {
        .data = {
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x34d869] # 0x3ddee0,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x90860,
            0x48, 0x83, 0xff, 0xbf, // cmp rdi, 0xffffffffffffffbf,
            0x0f, 0x87, 0xdf, 0x01, 0x00, 0x00, // ja 0x90870
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.34-1.2.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x10, // sub rsp, 0x10,
            0x80, 0x3d, 0x82, 0x4a, 0x16, 0x00, 0x00, // cmp BYTE PTR [rip+0x164a82], 0x0 # 0x201fe8,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x9d740,
            0x48, 0x85, 0xed, // test rbp, rbp,
            0x00, 0x00, // js 0x9d74e
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff // len=33
        }
    },

    // musl_1.2.2-4_amd64.so
    {
        .data = {
            0xe9, 0x8b, 0x19, 0x00, 0x00, // jmp 0x29400,
            0x66, 0x2e, 0x0f, 0x1f, 0x84, 0x00, 0x00, // cs nop WORD PTR [rax+rax*1+0x0],
            0x00, 0x00, 0x00, // ,
            0x90, // nop,
            0x55, // push rbp,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x53, // push rbx,
            0x48, 0xf7, 0xd8, // neg rax,
            0x48, 0x21, 0xf8, // and rax, rdi,
            0x48, 0x83, 0xec, 0x08, // sub rsp, 0x8,
            0x48, 0x39, // cmp rax, rdi
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0xff, 0xff, 0xff,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // musl-1.1.4-r5.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x28, // sub rsp, 0x28,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x18], rdi,
            0x48, 0x8d, 0x7c, 0x24, 0x18, // lea rdi, [rsp+0x18],
            0xe8, 0xef, 0x00, 0x00, 0x00, // call 0x25082,
            0x85, 0xc0, // test eax, eax,
            0x79, 0x07, // jns 0x2589e
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

};


static const uint FUNCSIGS_MALLOC_COUNT = 56;

static const funcsig FUNCSIGS_CALLOC[] = {

    // test-static2
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x55, // push r13,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x08, // sub rsp, 0x8,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x49, 0x89, 0xc4, // mov r12, rax,
            0x0f, 0x90, 0xc0, // seto al,
            0x4d, 0x85, 0xe4, // test r12, r12,
            0x0f, 0x88, 0x00, 0x00, // js 0x18018
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff // len=33
        }
    },

    // libc-2.33-5.fc34.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x55, // push r13,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x08, // sub rsp, 0x8,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x49, 0x89, 0xc4, // mov r12, rax,
            0x0f, 0x90, 0xc0, // seto al,
            0x4d, 0x85, 0xe4, // test r12, r12,
            0x0f, 0x88, 0x00, 0x00, // js 0x8d348
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff // len=33
        }
    },

    // libc6-amd64_2.24-3ubuntu1_i386.so
    {
        .data = {
            0x48, 0x89, 0xfa, // mov rdx, rdi,
            0x41, 0x56, // push r14,
            0xb8, 0xff, 0xff, 0xff, 0xff, // mov eax, 0xffffffff,
            0x41, 0x55, // push r13,
            0x48, 0x09, 0xf2, // or rdx, rsi,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x55, // push rbp,
            0x53, // push rbx,
            0x4c, 0x0f, 0xaf, 0xe6, // imul r12, rsi,
            0x48, 0x39, 0xc2, // cmp rdx, rax,
            0x76, 0x19, // jbe 0x7b818,
            0x48, 0x85, // test rsi, rsi
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // libc6_2.13-20ubuntu5.2_amd64.so
    {
        .data = {
            0x48, 0x89, 0xf2, // mov rdx, rsi,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x8], r15,
            0xb8, 0xff, 0xff, 0xff, 0xff, // mov eax, 0xffffffff,
            0x49, 0x89, 0xf7, // mov r15, rsi,
            0x48, 0x09, 0xfa, // or rdx, rdi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x30], rbx,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x28], rbp,
            0x4c, 0x89, 0x00, 0x00, // mov QWORD PTR [rsp-0x20], r12
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff // len=33
        }
    },

    // libc6_2.17-93ubuntu2_amd64.so
    {
        .data = {
            0x48, 0x89, 0xf2, // mov rdx, rsi,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x10], r13,
            0xb8, 0xff, 0xff, 0xff, 0xff, // mov eax, 0xffffffff,
            0x49, 0x89, 0xf5, // mov r13, rsi,
            0x48, 0x09, 0xfa, // or rdx, rdi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x28], rbx,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x20], rbp,
            0x4c, 0x89, 0x00, 0x00, // mov QWORD PTR [rsp-0x18], r12
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff // len=33
        }
    },

    // musl-1.1.12-r8.so
    {
        .data = {
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x263eb,
            0x48, 0x83, 0xc8, 0xff, // or rax, 0xffffffffffffffff,
            0x31, 0xd2, // xor edx, edx,
            0x48, 0xf7, 0xf6, // div rsi,
            0x48, 0x39, 0xf8, // cmp rax, rdi,
            0x72, 0x09, // jb 0x263f4,
            0x48, 0x0f, 0xaf, 0xfe, // imul rdi, rsi,
            0xe9, 0x56, 0x0c, 0x00, 0x00, // jmp 0x2704a,
            0x50, // push rax,
            0xe8, 0xad, 0x00, 0x00, // call 0x1daa7
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0xff, 0xff // len=33
        }
    },

    // libc6_2.7-9ubuntu2_amd64.so
    {
        .data = {
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x18], r12,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x28], rbx,
            0x48, 0x89, 0xf2, // mov rdx, rsi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x20], rbp,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x10], r13,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x8], r14,
            0x48, 0x83, // sub rsp, 0x28
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // libc6_2.19-13ubuntu3_amd64.so
    {
        .data = {
            0x41, 0x55, // push r13,
            0x48, 0x89, 0xfa, // mov rdx, rdi,
            0xb8, 0xff, 0xff, 0xff, 0xff, // mov eax, 0xffffffff,
            0x48, 0x09, 0xf2, // or rdx, rsi,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x4c, 0x0f, 0xaf, 0xe6, // imul r12, rsi,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x08, // sub rsp, 0x8,
            0x48, 0x39, 0xc2, // cmp rdx, rax,
            0x0f, 0x87, // ja 0x84005
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // libc6_2.31-17_amd64.so
    {
        .data = {
            0x41, 0x55, // push r13,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x08, // sub rsp, 0x8,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x49, 0x89, 0xc4, // mov r12, rax,
            0x0f, 0x90, 0xc0, // seto al,
            0x4d, 0x85, 0xe4, // test r12, r12,
            0x0f, 0x88, 0x00, 0x00, 0x00, 0x00, // js 0x8b038,
            0x0f, 0xb6, // movzx eax, al
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // libc-2.26-lp150.11.9.1.x86_64.so
    {
        .data = {
            0x41, 0x55, // push r13,
            0x48, 0x89, 0xfa, // mov rdx, rdi,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x48, 0x09, 0xf2, // or rdx, rsi,
            0xb8, 0xff, 0xff, 0xff, 0xff, // mov eax, 0xffffffff,
            0x48, 0x83, 0xec, 0x08, // sub rsp, 0x8,
            0x48, 0x0f, 0xaf, 0xee, // imul rbp, rsi,
            0x48, 0x39, 0xc2, // cmp rdx, rax,
            0x76, 0x09, // jbe 0x8556a
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // libc6_2.13-0ubuntu15_amd64.so
    {
        .data = {
            0x48, 0x89, 0xf2, // mov rdx, rsi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x28], rbp,
            0xb8, 0xff, 0xff, 0xff, 0xff, // mov eax, 0xffffffff,
            0x48, 0x89, 0xf5, // mov rbp, rsi,
            0x48, 0x09, 0xfa, // or rdx, rdi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x30], rbx,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x20], r12,
            0x4c, 0x89, 0x00, 0x00, // mov QWORD PTR [rsp-0x18], r13
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff // len=33
        }
    },

    // musl-1.2.2-3-omv4050.x86_64.so
    {
        .data = {
            0x41, 0x56, // push r14,
            0x53, // push rbx,
            0x50, // push rax,
            0x48, 0x89, 0xf3, // mov rbx, rsi,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x89, 0x00, 0x00, // mov QWORD PTR [rsp], rax,
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x5e445,
            0x48, 0x89, 0xd8, // mov rax, rbx,
            0x48, 0xf7, 0xe7, // mul rdi,
            0x71, 0x10, // jno 0x5e445
        }, 
        .undef = {
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // musl-1.1.22-r4.so
    {
        .data = {
            0x55, // push rbp,
            0x53, // push rbx,
            0x41, 0x50, // push r8,
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x23156,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x71, 0x12, // jno 0x23156,
            0xe8, 0x62, 0x00, 0x00, 0x00, // call 0x1c6ab,
            0x31, 0xed, // xor ebp, ebp,
            0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR [rax], 0xc,
            0xe9, 0x83, 0x00, // jmp 0x231d9
        }, 
        .undef = {
            0x00,
            0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.34-0ubuntu1_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x55, // push r13,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x49, 0x89, 0xc4, // mov r12, rax,
            0x0f, 0x90, 0xc0, // seto al,
            0x4d, 0x85, 0xe4, // test r12, r12,
            0x0f, 0x88, 0x00, 0x00, // js 0xa9c98
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff // len=33
        }
    },

    // libc6_2.29-0ubuntu2_amd64.so
    {
        .data = {
            0x48, 0x89, 0xfa, // mov rdx, rdi,
            0x41, 0x56, // push r14,
            0xb8, 0xff, 0xff, 0xff, 0xff, // mov eax, 0xffffffff,
            0x41, 0x55, // push r13,
            0x48, 0x09, 0xf2, // or rdx, rsi,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x0f, 0xaf, 0xee, // imul rbp, rsi,
            0x48, 0x39, 0xc2, // cmp rdx, rax,
            0x76, 0x09, // jbe 0x99a28,
            0x48, 0x85, // test rsi, rsi
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // libc6-amd64_2.4-1ubuntu12_i386.so
    {
        .data = {
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x18], r12,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x28], rbx,
            0x48, 0x89, 0xfa, // mov rdx, rdi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x20], rbp,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x10], r13,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x8], r14,
            0x48, 0x83, // sub rsp, 0x28
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // libc6-amd64_2.23-0ubuntu3_i386.so
    {
        .data = {
            0x41, 0x55, // push r13,
            0x48, 0x89, 0xfa, // mov rdx, rdi,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x48, 0x09, 0xf2, // or rdx, rsi,
            0xb8, 0xff, 0xff, 0xff, 0xff, // mov eax, 0xffffffff,
            0x48, 0x83, 0xec, 0x08, // sub rsp, 0x8,
            0x48, 0x0f, 0xaf, 0xee, // imul rbp, rsi,
            0x48, 0x39, 0xc2, // cmp rdx, rax,
            0x76, 0x1f, // jbe 0x79060
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // libc6_2.3.2.ds1-13ubuntu2.2_amd64_2.so
    {
        .data = {
            0x41, 0x56, // push r14,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rdx, QWORD PTR [rip+0x1c4157] # 0x23b5b0,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x41, 0x55, // push r13,
            0x48, 0x09, 0xf0, // or rax, rsi,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x55, // push rbp,
            0x53, // push rbx,
            0xbb, 0xff, 0xff, 0xff, 0xff, // mov ebx, 0xffffffff,
            0x48, 0x8b, 0x00, // mov rcx, QWORD PTR [rdx],
            0x4c, // imul r12, rsi
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff,
            0x00 // len=33
        }
    },

    // musl-1.1.24-r3.so
    {
        .data = {
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x41, 0x50, // push r8,
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x23523,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x71, 0x13, // jno 0x23523,
            0xe8, 0x92, 0x00, 0x00, 0x00, // call 0x1c6a7,
            0x45, 0x31, 0xed, // xor r13d, r13d,
            0xc7, 0x00, 0x0c, // mov DWORD PTR [rax], 0xc
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // musl-1.2.2-r7.so
    {
        .data = {
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x41, 0x50, // push r8,
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x23327,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x71, 0x13, // jno 0x23327,
            0xe8, 0x9a, 0x00, 0x00, 0x00, // call 0x1ceb3,
            0x45, 0x31, 0xed, // xor r13d, r13d,
            0xc7, 0x00, 0x0c, // mov DWORD PTR [rax], 0xc
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // glibc-2.28-160.el8.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xfa, // mov rdx, rdi,
            0x41, 0x56, // push r14,
            0xb8, 0xff, 0xff, 0xff, 0xff, // mov eax, 0xffffffff,
            0x41, 0x55, // push r13,
            0x48, 0x09, 0xf2, // or rdx, rsi,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x0f, 0xaf, 0xee, // imul rbp, rsi,
            0x48, 0x39, 0xc2, // cmp rdx, rax
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.18-0ubuntu1_amd64.so
    {
        .data = {
            0x41, 0x56, // push r14,
            0x48, 0x89, 0xfa, // mov rdx, rdi,
            0xb8, 0xff, 0xff, 0xff, 0xff, // mov eax, 0xffffffff,
            0x48, 0x09, 0xf2, // or rdx, rsi,
            0x41, 0x55, // push r13,
            0x49, 0x89, 0xfd, // mov r13, rdi,
            0x4c, 0x0f, 0xaf, 0xee, // imul r13, rsi,
            0x48, 0x39, 0xc2, // cmp rdx, rax,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x0f, 0x87, 0x81, 0x02, // ja 0x83bc4
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // musl_1.2.1-1_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x48, 0x89, 0xf7, // mov rdi, rsi,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x2a34f,
            0x4c, 0x89, 0xe0, // mov rax, r12,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x0f, 0x80, 0xb2, 0x00, 0x00, 0x00, // jo 0x2a401,
            0x4c, 0x0f, // imul r12, rdi
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // libc6_2.3.2.ds1-13ubuntu2.2_amd64.so
    {
        .data = {
            0x41, 0x56, // push r14,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1bff27] # 0x2340d0,
            0x48, 0x89, 0xfa, // mov rdx, rdi,
            0x41, 0x55, // push r13,
            0x48, 0x09, 0xf2, // or rdx, rsi,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x8b, 0x00, // mov rcx, QWORD PTR [rax],
            0xb8, 0xff, 0xff, 0xff, 0xff, // mov eax, 0xffffffff,
            0x4c, // imul r12, rsi
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc-2.34-11.mga9.x86_64_3.so
    {
        .data = {
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x41, 0x55, // push r13,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x41, 0x54, // push r12,
            0x53, // push rbx,
            0x0f, 0x80, 0x87, 0x01, 0x00, 0x00, // jo 0x6a48,
            0x49, 0x89, 0xc5, // mov r13, rax,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x36dd] # 0x9fa8,
            0x48, 0x8b, 0x00, // mov rcx, QWORD PTR [rax],
            0x48, 0x85, 0xc9, // test rcx, rcx
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // musl-1.1.11-r5.so
    {
        .data = {
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x26509,
            0x48, 0x83, 0xc8, 0xff, // or rax, 0xffffffffffffffff,
            0x31, 0xd2, // xor edx, edx,
            0x48, 0xf7, 0xf6, // div rsi,
            0x48, 0x39, 0xc7, // cmp rdi, rax,
            0x77, 0x09, // ja 0x26512,
            0x48, 0x0f, 0xaf, 0xfe, // imul rdi, rsi,
            0xe9, 0x74, 0x0c, 0x00, 0x00, // jmp 0x27186,
            0x50, // push rax,
            0xe8, 0xb5, 0x00, 0x00, // call 0x1dbcd
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0xff, 0xff // len=33
        }
    },

    // musl-1.1.14-r16.so
    {
        .data = {
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x2345a,
            0x48, 0x83, 0xc8, 0xff, // or rax, 0xffffffffffffffff,
            0x31, 0xd2, // xor edx, edx,
            0x48, 0xf7, 0xf6, // div rsi,
            0x48, 0x39, 0xf8, // cmp rax, rdi,
            0x72, 0x09, // jb 0x23463,
            0x48, 0x0f, 0xaf, 0xfe, // imul rdi, rsi,
            0xe9, 0x76, 0x0b, 0x00, 0x00, // jmp 0x23fd9,
            0x50, // push rax,
            0xe8, 0x21, 0x00, 0x00, // call 0x1dd8a
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0xff, 0xff // len=33
        }
    },

    // libc-2.17-317.el7.x86_64.so
    {
        .data = {
            0x41, 0x56, // push r14,
            0x48, 0x89, 0xfa, // mov rdx, rdi,
            0xb8, 0xff, 0xff, 0xff, 0xff, // mov eax, 0xffffffff,
            0x48, 0x09, 0xf2, // or rdx, rsi,
            0x41, 0x55, // push r13,
            0x49, 0x89, 0xfd, // mov r13, rdi,
            0x4c, 0x0f, 0xaf, 0xee, // imul r13, rsi,
            0x48, 0x39, 0xc2, // cmp rdx, rax,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x0f, 0x87, 0x48, 0x02, // ja 0x8633b
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.34-0-omv4050.x86_64.so
    {
        .data = {
            0x41, 0x55, // push r13,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x49, 0x89, 0xc4, // mov r12, rax,
            0x0f, 0x90, 0xc0, // seto al,
            0x4d, 0x85, 0xe4, // test r12, r12,
            0x0f, 0x88, 0x00, 0x00, 0x00, 0x00, // js 0xa5e70,
            0x0f, 0xb6, // movzx eax, al
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // musl_1.1.24-1_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x2b488,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x70, 0x3e, // jo 0x2b4c6,
            0x4c, 0x0f, 0xaf, 0xe6, // imul r12, rsi,
            0x4c, 0x89, 0xe7, // mov rdi, r12,
            0x00, 0x00, // call 0x15010
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff // len=33
        }
    },

    // libc6-amd64_2.9-4ubuntu6.3_i386.so
    {
        .data = {
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x20], r12,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x30], rbx,
            0x48, 0x89, 0xf2, // mov rdx, rsi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x28], rbp,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x18], r13,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x10], r14,
            0x00, 0x00, // mov QWORD PTR [rsp-0x8], r15
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0xff, 0xff // len=33
        }
    },

    // libc-2.29-3-x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xfa, // mov rdx, rdi,
            0x41, 0x56, // push r14,
            0xb8, 0xff, 0xff, 0xff, 0xff, // mov eax, 0xffffffff,
            0x41, 0x55, // push r13,
            0x48, 0x09, 0xf2, // or rdx, rsi,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x55, // push rbp,
            0x4c, 0x0f, 0xaf, 0xe6, // imul r12, rsi,
            0x53, // push rbx,
            0x48, 0x39, 0xc2, // cmp rdx, rax
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6-amd64_2.3.6-0ubuntu20.6_i386.so
    {
        .data = {
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x28], rbx,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x18], r12,
            0x48, 0x89, 0xfa, // mov rdx, rdi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x20], rbp,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x10], r13,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x8], r14,
            0x48, 0x83, // sub rsp, 0x28
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // musl-1.1.22-2-x86_64.so
    {
        .data = {
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x18741,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x70, 0x43, // jo 0x18784,
            0x48, 0x89, 0xf3, // mov rbx, rsi,
            0x48, 0x0f, 0xaf, 0xdf, // imul rbx, rdi,
            0x48, 0x89, 0xdf, // mov rdi, rbx,
            0xe8, 0xc0, 0x00, 0x00, 0x00, // call 0x15010,
            0x49, // mov r12, rax
        }, 
        .undef = {
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // musl_0.9.15-1_amd64.so
    {
        .data = {
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x53, // push rbx,
            0x00, 0x00, // je 0x23337,
            0x31, 0xd2, // xor edx, edx,
            0x48, 0xc7, 0xc0, 0xff, 0xff, 0xff, 0xff, // mov rax, 0xffffffffffffffff,
            0x48, 0xf7, 0xf6, // div rsi,
            0x48, 0x39, 0xf8, // cmp rax, rdi,
            0x72, 0x51, // jb 0x23388,
            0x48, 0x89, 0xf3, // mov rbx, rsi,
            0x48, 0x0f, 0xaf, 0xdf, // imul rbx, rdi,
            0x48, 0x89, 0xdf, // mov rdi, rbx
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00,
            0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.34-0ubuntu2_amd64_2.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x30, // sub rsp, 0x30,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x28], rax,
            0x48, 0x89, 0xf8, // mov rax, rdi
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // musl-1.1.24-r10.so
    {
        .data = {
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x41, 0x50, // push r8,
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x23578,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x71, 0x13, // jno 0x23578,
            0xe8, 0x37, 0x00, 0x00, 0x00, // call 0x1c6a1,
            0x45, 0x31, 0xed, // xor r13d, r13d,
            0xc7, 0x00, 0x0c, // mov DWORD PTR [rax], 0xc
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // musl-1.1.23-1-x86_64.so
    {
        .data = {
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x18724,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x70, 0x3e, // jo 0x18762,
            0x4c, 0x0f, 0xaf, 0xe6, // imul r12, rsi,
            0x4c, 0x89, 0xe7, // mov rdi, r12,
            0xe8, 0xe0, 0x00, 0x00, 0x00, // call 0x15010,
            0x48, // mov rbp, rax
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // libc6-amd64_2.34-0ubuntu3_i386_2.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x41, 0x55, // push r13,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x41, 0x54, // push r12,
            0x53, // push rbx,
            0x0f, 0x80, 0x73, 0x01, 0x00, 0x00, // jo 0x70e8,
            0x49, 0x89, 0xc5, // mov r13, rax,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x5029] # 0xbfa8,
            0x00, 0x00, // mov rdx, QWORD PTR [rax]
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff // len=33
        }
    },

    // musl-1.1.19-r11.so
    {
        .data = {
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x23c76,
            0x48, 0x83, 0xc8, 0xff, // or rax, 0xffffffffffffffff,
            0x31, 0xd2, // xor edx, edx,
            0x48, 0xf7, 0xf6, // div rsi,
            0x48, 0x39, 0xf8, // cmp rax, rdi,
            0x72, 0x09, // jb 0x23c7f,
            0x48, 0x0f, 0xaf, 0xfe, // imul rdi, rsi,
            0xe9, 0x9b, 0x0b, 0x00, 0x00, // jmp 0x2481a,
            0x48, 0x83, 0xec, 0x08, // sub rsp, 0x8,
            0x00, // call 0x1df31
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0xff // len=33
        }
    },

    // musl_1.1.21-2_amd64.so
    {
        .data = {
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x2a2d1,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x70, 0x43, // jo 0x2a314,
            0x48, 0x89, 0xf3, // mov rbx, rsi,
            0x48, 0x0f, 0xaf, 0xdf, // imul rbx, rdi,
            0x48, 0x89, 0xdf, // mov rdi, rbx,
            0xe8, 0x30, 0x00, 0x00, 0x00, // call 0x15010,
            0x49, // mov r12, rax
        }, 
        .undef = {
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // libc-2.31-7.30.x86_64.so
    {
        .data = {
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x08, // sub rsp, 0x8,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x48, 0x89, 0xc3, // mov rbx, rax,
            0x0f, 0x90, 0xc0, // seto al,
            0x48, 0x85, 0xdb, // test rbx, rbx,
            0x0f, 0x88, 0x00, 0x00, 0x00, 0x00, // js 0x8aee0,
            0x0f, 0xb6, // movzx eax, al
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // musl-1.2.2-1-x86_64.so
    {
        .data = {
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x48, 0x89, 0xf7, // mov rdi, rsi,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x1750b,
            0x4c, 0x89, 0xe0, // mov rax, r12,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x0f, 0x80, 0xb6, 0x00, 0x00, 0x00, // jo 0x175c1,
            0x4c, 0x0f, 0xaf, 0xe7, // imul r12, rdi,
            0x4c, 0x89, // mov rdi, r12
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // musl-1.1.5-r5.so
    {
        .data = {
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x53, // push rbx,
            0x00, 0x00, // je 0x2593b,
            0x48, 0x83, 0xc8, 0xff, // or rax, 0xffffffffffffffff,
            0x31, 0xd2, // xor edx, edx,
            0x48, 0xf7, 0xf6, // div rsi,
            0x48, 0x39, 0xc7, // cmp rdi, rax,
            0x76, 0x0d, // jbe 0x2593b,
            0xe8, 0x9a, 0x00, 0x00, 0x00, // call 0x1d1cd,
            0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR [rax], 0xc,
            0xeb, 0x45, // jmp 0x25980
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // musl_1.1.19-1_amd64.so
    {
        .data = {
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x29f1d,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x70, 0x09, // jo 0x29f26,
            0x48, 0x0f, 0xaf, 0xfe, // imul rdi, rsi,
            0xe9, 0xea, 0x11, 0x00, 0x00, // jmp 0x2b110,
            0x48, 0x83, 0xec, 0x08, // sub rsp, 0x8,
            0xe8, 0x11, 0x00, 0x00, 0x00, // call 0x20740,
            0xc7, 0x00, // mov DWORD PTR [rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // musl_1.1.16-3_amd64.so
    {
        .data = {
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x28216,
            0x31, 0xd2, // xor edx, edx,
            0x48, 0xc7, 0xc0, 0xff, 0xff, 0xff, 0xff, // mov rax, 0xffffffffffffffff,
            0x48, 0xf7, 0xf6, // div rsi,
            0x48, 0x39, 0xf8, // cmp rax, rdi,
            0x72, 0x0a, // jb 0x28220,
            0x48, 0x0f, 0xaf, 0xfe, // imul rdi, rsi,
            0xe9, 0x91, 0x11, 0x00, 0x00, // jmp 0x293b0,
            0x90, // nop,
            0x48, // sub rsp, 0x8
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00 // len=33
        }
    },

    // musl-1.1.20-r6.so
    {
        .data = {
            0x55, // push rbp,
            0x53, // push rbx,
            0x41, 0x50, // push r8,
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x24d71,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x71, 0x12, // jno 0x24d71,
            0xe8, 0x23, 0x00, 0x00, 0x00, // call 0x1e587,
            0x31, 0xed, // xor ebp, ebp,
            0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR [rax], 0xc,
            0xe9, 0x83, 0x00, // jmp 0x24df4
        }, 
        .undef = {
            0x00,
            0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.34-0-omv4050.x86_64_2.so
    {
        .data = {
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x30, // sub rsp, 0x30,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x28], rax,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x0f, // jo 0x72e7
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00 // len=33
        }
    },

    // musl-1.1.15-r8.so
    {
        .data = {
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x235f0,
            0x48, 0x83, 0xc8, 0xff, // or rax, 0xffffffffffffffff,
            0x31, 0xd2, // xor edx, edx,
            0x48, 0xf7, 0xf6, // div rsi,
            0x48, 0x39, 0xf8, // cmp rax, rdi,
            0x72, 0x09, // jb 0x235f9,
            0x48, 0x0f, 0xaf, 0xfe, // imul rdi, rsi,
            0xe9, 0x72, 0x0b, 0x00, 0x00, // jmp 0x2416b,
            0x48, 0x83, 0xec, 0x08, // sub rsp, 0x8,
            0x00, // call 0x1de97
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0xff // len=33
        }
    },

    // musl-1.1.4-r5.so
    {
        .data = {
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x53, // push rbx,
            0x00, 0x00, // je 0x24ecf,
            0x48, 0x83, 0xc8, 0xff, // or rax, 0xffffffffffffffff,
            0x31, 0xd2, // xor edx, edx,
            0x48, 0xf7, 0xf6, // div rsi,
            0x48, 0x39, 0xc7, // cmp rdi, rax,
            0x76, 0x0d, // jbe 0x24ecf,
            0xe8, 0xe8, 0x00, 0x00, 0x00, // call 0x1c7af,
            0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR [rax], 0xc,
            0xeb, 0x45, // jmp 0x24f14
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // musl-1.2.2-r3.so
    {
        .data = {
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x41, 0x50, // push r8,
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x23328,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x71, 0x13, // jno 0x23328,
            0xe8, 0x99, 0x00, 0x00, 0x00, // call 0x1ceb3,
            0x45, 0x31, 0xed, // xor r13d, r13d,
            0xc7, 0x00, 0x0c, // mov DWORD PTR [rax], 0xc
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00 // len=33
        }
    },

};


static const uint FUNCSIGS_CALLOC_COUNT = 51;

static const funcsig FUNCSIGS_FREE[] = {

    // test-static2
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x99ec7] # 0xb1218,
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x173e0,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x00, 0x00, // je 0x173a6,
            0x48, 0xc7, // mov rbx, 0xffffffffffffffc0
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // libc-2.33-5.fc34.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1364df] # 0x1c2ee0,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x8ca98,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x00, // je 0x8ca5b
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff // len=33
        }
    },

    // libc6-amd64_2.24-3ubuntu1_i386.so
    {
        .data = {
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x31db01] # 0x398ef8,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x00, 0x00, // jne 0x7b478,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x00, 0x00, // je 0x7b480,
            0x48, 0x8b, 0x00, 0x00, // mov rax, QWORD PTR [rdi-0x8],
            0x48, 0x8d, 0x77, 0xf0, // lea rsi, [rdi-0x10],
            0xa8, 0x02, // test al, 0x2,
            0x00, 0x00, // jne 0x7b430,
            0xa8, // test al, 0x4
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0xff, 0xff,
            0x00 // len=33
        }
    },

    // libc6_2.13-20ubuntu5.2_amd64.so
    {
        .data = {
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x31ce09] # 0x39bea0,
            0x53, // push rbx,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x7f159,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x00, 0x00, // je 0x7f118,
            0x48, 0x8b, 0x00, 0x00, // mov rax, QWORD PTR [rdi-0x8],
            0x48, 0x8d, 0x57, 0xf0, // lea rdx, [rdi-0x10]
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.17-93ubuntu2_amd64.so
    {
        .data = {
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x33d5e1] # 0x3c0ee8,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x839d2,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x839d0,
            0x48, 0x8b, 0x00, 0x00, // mov rax, QWORD PTR [rdi-0x8],
            0x48, // lea rsi, [rdi-0x10]
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // libc6_2.32-0ubuntu3.2_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x14b359] # 0x1e2ed8,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x97c08,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x00, 0x00, // je 0x97c00,
            0x00, // mov rax, QWORD PTR [rdi-0x8]
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0xff // len=33
        }
    },

    // musl-1.1.12-r8.so
    {
        .data = {
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x26c46,
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x4c, 0x8d, 0x7f, 0xf0, // lea r15, [rdi-0x10],
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x28, // sub rsp, 0x28,
            0x48, 0x8b, 0x00, 0x00, // mov rsi, QWORD PTR [rdi-0x8],
            0x40, 0xf6, // test sil, 0x1
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // libc6_2.7-9ubuntu2_amd64.so
    {
        .data = {
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x10], rbx,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x8], rbp,
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x2e436b] # 0x35bf60,
            0x48, 0x89, 0xfa, // mov rdx, rdi,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x00, 0x00, // je 0x77c16,
            0x00, // mov rsi, QWORD PTR [rsp+0x38]
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0xff // len=33
        }
    },

    // libc6_2.31-17_amd64.so
    {
        .data = {
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1337cd] # 0x1bdef8,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x8a7b8,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x00, 0x00, // je 0x8a7b0,
            0x48, 0x8b, 0x00, 0x00, // mov rax, QWORD PTR [rdi-0x8],
            0x48, // lea rsi, [rdi-0x10]
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // libc-2.26-lp150.11.9.1.x86_64.so
    {
        .data = {
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x10, // sub rsp, 0x10,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x32ef8c] # 0x3b3ef8,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x85020,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x85010
        }, 
        .undef = {
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff // len=33
        }
    },

    // libc6_2.13-0ubuntu15_amd64.so
    {
        .data = {
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x20, // sub rsp, 0x20,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x32276c] # 0x39ef68,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x7c970,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x00, 0x00, // je 0x7c87f,
            0x48, 0x8b, 0x00, 0x00, // mov rsi, QWORD PTR [rdi-0x8]
        }, 
        .undef = {
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff // len=33
        }
    },

    // libc6_2.24-7ubuntu2_amd64.so
    {
        .data = {
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x28, // sub rsp, 0x28,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x33ada7] # 0x3c0ef8,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x86260,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x00, // je 0x8618c
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff // len=33
        }
    },

    // musl-1.2.2-3-omv4050.x86_64.so
    {
        .data = {
            0x50, // push rax,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x89, 0x00, 0x00, // mov QWORD PTR [rsp], rax,
            0xe8, 0xed, 0x00, 0x00, 0x00, // call 0x5eb3b,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x3b, 0x00, 0x00, // cmp rax, QWORD PTR [rsp],
            0x00, // jne 0x5e55f
        }, 
        .undef = {
            0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0xff // len=33
        }
    },

    // musl-1.1.22-r4.so
    {
        .data = {
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x00, 0x00, // je 0x23217,
            0x48, 0x8b, 0x00, 0x00, // mov rsi, QWORD PTR [rdi-0x8],
            0x48, 0x8d, 0x47, 0xf0, // lea rax, [rdi-0x10],
            0x40, 0xf6, 0xc6, 0x01, // test sil, 0x1,
            0x00, 0x00, // jne 0x2320f,
            0x48, 0x8b, 0x00, 0x00, // mov rdx, QWORD PTR [rdi-0x10],
            0x48, 0x83, 0xe6, 0xfe, // and rsi, 0xfffffffffffffffe,
            0x48, 0x29, 0xd0, // sub rax, rdx,
            0x48, 0x01, 0xd6, // add rsi, rdx
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.34-0ubuntu1_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0xa8b08,
            0x55, // push rbp,
            0x48, 0x8d, 0x77, 0xf0, // lea rsi, [rdi-0x10],
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rbx, QWORD PTR [rip+0x16f392] # 0x217e10,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rdi-0x8]
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff // len=33
        }
    },

    // libc6_2.29-0ubuntu2_amd64.so
    {
        .data = {
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x10, // sub rsp, 0x10,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x14ad1c] # 0x1e3ef8,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x99270,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x00, 0x00, // je 0x99260,
            0x48, 0x8b, 0x00, 0x00, // mov rax, QWORD PTR [rdi-0x8]
        }, 
        .undef = {
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff // len=33
        }
    },

    // libc6_2.27-5ubuntu1_amd64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x28], rax,
            0x31, 0xc0, // xor eax, eax
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // libc6-amd64_2.4-1ubuntu12_i386.so
    {
        .data = {
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x10], rbx,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x8], rbp,
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1cb6f3] # 0x239f38,
            0x49, 0x89, 0xf8, // mov r8, rdi,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x00, 0x00, // je 0x6e866,
            0x00, // mov rsi, QWORD PTR [rsp+0x38]
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0xff // len=33
        }
    },

    // libc6_2.3.2.ds1-13ubuntu2.2_amd64_2.so
    {
        .data = {
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rcx, QWORD PTR [rip+0x1c4971] # 0x23b828,
            0x53, // push rbx,
            0x48, 0x89, 0xfa, // mov rdx, rdi,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rcx],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x00, 0x00, // je 0x76ecc,
            0x48, 0x8b, 0x00, 0x00, 0x00, // mov rsi, QWORD PTR [rsp+0x8],
            0x00, 0x00, // call rax,
            0x5b, // pop rbx,
            0xc3, // ret,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x00, 0x00, // je 0x76eca
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff // len=33
        }
    },

    // musl-1.1.24-r3.so
    {
        .data = {
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x00, 0x00, // je 0x235e4,
            0x48, 0x8b, 0x00, 0x00, // mov rsi, QWORD PTR [rdi-0x8],
            0x4c, 0x8d, 0x47, 0xf0, // lea r8, [rdi-0x10],
            0x40, 0xf6, 0xc6, 0x01, // test sil, 0x1,
            0x00, 0x00, // jne 0x235dc,
            0x48, 0x8b, 0x00, 0x00, // mov rax, QWORD PTR [rdi-0x10],
            0x48, 0x83, 0xe6, 0xfe, // and rsi, 0xfffffffffffffffe,
            0x4c, 0x89, 0xc7, // mov rdi, r8,
            0x48, 0x29, 0xc7, // sub rdi, rax
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // musl-1.2.2-r7.so
    {
        .data = {
            0xe9, 0x54, 0x08, 0x00, 0x00, // jmp 0x23c16,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x41, 0x50, // push r8,
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x233eb,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x71, 0x13, // jno 0x233eb,
            0xe8, 0xd6, 0x00, 0x00, 0x00, // call 0x1ceb3,
            0x45, // xor r13d, r13d
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // glibc-2.28-160.el8.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x10, // sub rsp, 0x10,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x338e00] # 0x3beef0,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x86180,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x00, 0x00, // je 0x86170
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff // len=33
        }
    },

    // musl_1.2.1-1_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x2af48,
            0x41, 0x55, // push r13,
            0x49, 0x89, 0xfa, // mov r10, rdi,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x08, // sub rsp, 0x8,
            0xe8, 0xd1, 0x00, 0x00, 0x00, // call 0x2a8d0,
            0x41, 0x0f, // movzx edx, BYTE PTR [r10-0x3]
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // libc6_2.3.2.ds1-13ubuntu2.2_amd64.so
    {
        .data = {
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1c080d] # 0x2343a8,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x8], rbx,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x10], rbp,
            0x48, 0x89, 0xfb, // mov rbx, rdi,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x00, 0x00, // je 0x73bc6,
            0x00, // mov rsi, QWORD PTR [rsp+0x18]
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0xff // len=33
        }
    },

    // libc-2.34-11.mga9.x86_64_3.so
    {
        .data = {
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x55, // push rbp,
            0x48, 0x83, 0xec, 0x20, // sub rsp, 0x20,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x4445] # 0x9fc8,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x5d28,
            0x8b, 0x05, // mov eax, DWORD PTR [rip+0x4f63] # 0xaaf8
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // musl-1.1.11-r5.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x4c, 0x8d, 0x7f, 0xf0, // lea r15, [rdi-0x10],
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x28, // sub rsp, 0x28,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x26d71,
            0x48, 0x8b, 0x00, 0x00, // mov rsi, QWORD PTR [rdi-0x8],
            0x40, 0xf6, // test sil, 0x1
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // libc-2.34-0-omv4050.x86_64.so
    {
        .data = {
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0xa5068,
            0x55, // push rbp,
            0x48, 0x8d, 0x77, 0xf0, // lea rsi, [rdi-0x10],
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rbx, QWORD PTR [rip+0x165e1e] # 0x20adf8,
            0x48, 0x8b, 0x00, 0x00, // mov rax, QWORD PTR [rdi-0x8],
            0x64, 0x8b, 0x2b, // mov ebp, DWORD PTR fs:[rbx]
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // musl_1.1.24-1_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x00, 0x00, // je 0x2b580,
            0x48, 0x8b, 0x00, 0x00, // mov rsi, QWORD PTR [rdi-0x8],
            0x4c, 0x8d, 0x47, 0xf0, // lea r8, [rdi-0x10],
            0x40, 0xf6, 0xc6, 0x01, // test sil, 0x1,
            0x00, 0x00, // jne 0x2b578,
            0x48, 0x8b, 0x00, 0x00, // mov rax, QWORD PTR [rdi-0x10],
            0x48, 0x83, 0xe6, 0xfe, // and rsi, 0xfffffffffffffffe,
            0x4c, 0x89, // mov rdi, r8
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // libc6-amd64_2.9-4ubuntu6.3_i386.so
    {
        .data = {
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x2f1d21] # 0x36bf58,
            0x53, // push rbx,
            0x49, 0x89, 0xf8, // mov r8, rdi,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x00, 0x00, // je 0x7a250,
            0x48, 0x8b, 0x00, 0x00, 0x00, // mov rsi, QWORD PTR [rsp+0x8],
            0x49, 0x89, 0xc3, // mov r11, rax,
            0x5b, // pop rbx,
            0x41, 0xff, 0xe3, // jmp r11,
            0x90, // nop,
            0x48, // test rdi, rdi
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00 // len=33
        }
    },

    // libc6_2.26-0ubuntu2_amd64.so
    {
        .data = {
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x28], rax,
            0x31, 0xc0, // xor eax, eax,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x34ab37] # 0x3d9ee8
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff // len=33
        }
    },

    // libc6-amd64_2.3.6-0ubuntu20.6_i386.so
    {
        .data = {
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1c92d9] # 0x232f30,
            0x53, // push rbx,
            0x48, 0x89, 0xfa, // mov rdx, rdi,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x00, 0x00, // je 0x69c70,
            0x48, 0x8b, 0x00, 0x00, 0x00, // mov rsi, QWORD PTR [rsp+0x8],
            0x49, 0x89, 0xc3, // mov r11, rax,
            0x5b, // pop rbx,
            0x41, 0xff, 0xe3, // jmp r11,
            0x90, // nop,
            0x48, // test rdi, rdi
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00 // len=33
        }
    },

    // musl_0.9.15-1_amd64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x4c, 0x8d, 0x67, 0xf0, // lea r12, [rdi-0x10],
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x48, // sub rsp, 0x48,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x23d60,
            0x48, 0x8b, 0x00, 0x00, // mov rax, QWORD PTR [rdi-0x8],
            0xa8, 0x01, // test al, 0x1
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // libc6_2.34-0ubuntu2_amd64_2.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x28], rax,
            0x00, 0x00, // mov rax, QWORD PTR [rip+0x5c52] # 0xbfc8
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0xff, 0xff // len=33
        }
    },

    // libc6_2.33-0experimental2_amd64.so
    {
        .data = {
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1325b3] # 0x1bcee0,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x8a9c0,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x00, 0x00, // je 0x8a987,
            0x48, 0x8b, 0x00, // mov rbx, QWORD PTR [rip+0x13250b] # 0x1bce50
        }, 
        .undef = {
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0xff // len=33
        }
    },

    // musl_1.1.5-2+deb8u1_amd64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x4c, 0x8d, 0x67, 0xf0, // lea r12, [rdi-0x10],
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x26b88,
            0x48, 0x8b, 0x00, 0x00, // mov rax, QWORD PTR [rdi-0x8],
            0xa8, 0x01, // test al, 0x1
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // musl-1.1.19-r11.so
    {
        .data = {
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x2440b,
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x4c, 0x8d, 0x67, 0xf0, // lea r12, [rdi-0x10],
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x28, // sub rsp, 0x28,
            0x48, 0x8b, 0x00, 0x00, // mov rsi, QWORD PTR [rdi-0x8],
            0x40, 0xf6, // test sil, 0x1
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // libc-2.34-11.mga9.x86_64.so
    {
        .data = {
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x96b20,
            0x41, 0x54, // push r12,
            0x48, 0x8d, 0x77, 0xf0, // lea rsi, [rdi-0x10],
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x10, // sub rsp, 0x10,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rbp, QWORD PTR [rip+0x14f394] # 0x1e5e10,
            0x48, 0x8b, 0x00, 0x00, // mov rax, QWORD PTR [rdi-0x8],
            0x00, // mov rbx, QWORD PTR fs:0x0
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0xff // len=33
        }
    },

    // libc-2.34-7.fc35.x86_64_2.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x55, // push rbp,
            0x48, 0x83, 0xec, 0x30, // sub rsp, 0x30,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x28], rax,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rip+0x5e13] # 0xbfc8
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff // len=33
        }
    },

    // musl-1.2.2-1-x86_64.so
    {
        .data = {
            0xe9, 0x93, 0xfa, 0x00, 0x00, // jmp 0x27078,
            0x66, 0x2e, 0x0f, 0x1f, 0x84, 0x00, 0x00, // cs nop WORD PTR [rax+rax*1+0x0],
            0x00, 0x00, 0x00, // ,
            0x90, // nop,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x48, 0x89, 0xf7, // mov rdi, rsi,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x1760b,
            0x4c, 0x89, // mov rax, r12
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0xff, 0xff, 0xff,
            0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // musl-1.1.5-r5.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x4c, 0x8d, 0x77, 0xf0, // lea r14, [rdi-0x10],
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x26296,
            0x48, 0x8b, 0x00, 0x00, // mov rsi, QWORD PTR [rdi-0x8],
            0x40, 0xf6, // test sil, 0x1
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // musl_1.1.19-1_amd64.so
    {
        .data = {
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x2aaf8,
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x8d, 0x6f, 0xf0, // lea rbp, [rdi-0x10],
            0x48, 0x83, 0xec, 0x28, // sub rsp, 0x28,
            0x48, 0x8b, 0x00, 0x00, // mov rax, QWORD PTR [rdi-0x8],
            0x48, 0x89, // mov rsi, rax
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // musl_1.1.16-3_amd64.so
    {
        .data = {
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x28dce,
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x4c, 0x8d, 0x67, 0xf0, // lea r12, [rdi-0x10],
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x48, 0x8b, 0x00, 0x00, // mov rsi, QWORD PTR [rdi-0x8],
            0x40, 0xf6, // test sil, 0x1
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // musl-1.2.1-2-x86_64.so
    {
        .data = {
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x26f84,
            0x41, 0x55, // push r13,
            0x49, 0x89, 0xf9, // mov r9, rdi,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x51, // push rcx,
            0xe8, 0x24, 0x00, 0x00, 0x00, // call 0x26a3e,
            0x45, 0x8a, 0x41, 0xfd, // mov r8b, BYTE PTR [r9-0x3],
            0x48, 0x89, 0xc7, // mov rdi, rax,
            0x49, 0x89, // mov r12, rax
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // musl-1.2.2-1-omv4002.x86_64.so
    {
        .data = {
            0x50, // push rax,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x89, 0x00, 0x00, // mov QWORD PTR [rsp], rax,
            0xe8, 0xd6, 0x00, 0x00, 0x00, // call 0x5eab9,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x3b, 0x00, 0x00, // cmp rax, QWORD PTR [rsp],
            0x00, // jne 0x5e4f4
        }, 
        .undef = {
            0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0xff // len=33
        }
    },

    // libc-2.34-0-omv4050.x86_64_2.so
    {
        .data = {
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x55, // push rbp,
            0x48, 0x83, 0xec, 0x30, // sub rsp, 0x30,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x28], rax,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x5077] # 0xafc8
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff // len=33
        }
    },

    // libc6_2.26-0ubuntu4_amd64.so
    {
        .data = {
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x34cfdd] # 0x3ddee8,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x910a0,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x91099,
            0x00, // mov rax, QWORD PTR [rdi-0x8]
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0xff // len=33
        }
    },

    // musl_1.2.2-4_amd64.so
    {
        .data = {
            0xe9, 0xcb, 0x0a, 0x00, 0x00, // jmp 0x28430,
            0x66, 0x2e, 0x0f, 0x1f, 0x84, 0x00, 0x00, // cs nop WORD PTR [rax+rax*1+0x0],
            0x00, 0x00, 0x00, // ,
            0x90, // nop,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x48, 0x89, 0xf7, // mov rdi, rsi,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x2798b,
            0x4c, 0x89, // mov rax, r12
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0xff, 0xff, 0xff,
            0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // musl-1.1.4-r5.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x8d, 0x5f, 0xf0, // lea rbx, [rdi-0x10],
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x25828,
            0x48, 0x8b, 0x00, 0x00, // mov rsi, QWORD PTR [rdi-0x8],
            0x40, 0xf6, // test sil, 0x1
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },

    // musl-1.2.2-r3.so
    {
        .data = {
            0xe9, 0x54, 0x08, 0x00, 0x00, // jmp 0x23c17,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x41, 0x50, // push r8,
            0x48, 0x85, 0xf6, // test rsi, rsi,
            0x00, 0x00, // je 0x233ec,
            0x48, 0x89, 0xf8, // mov rax, rdi,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x71, 0x13, // jno 0x233ec,
            0xe8, 0xd5, 0x00, 0x00, 0x00, // call 0x1ceb3,
            0x45, // xor r13d, r13d
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

};


static const uint FUNCSIGS_FREE_COUNT = 49;

static const funcsig FUNCSIGS_REALLOC[] = {
    
    // test-static2
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x971e9] # 0xae7c8,
            0x48, 0x85, // test rax, rax
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00 // len=33
        }
    },


    // libc-2.33-5.fc34.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x136321] # 0x1c2fd0,
            0x00, 0x00, // mov rax, QWORD PTR [rax]
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff // len=33
        }
    },

    // libc6-amd64_2.24-3ubuntu1_i386.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x31db1b] # 0x398fd0,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x0f, 0x85, 0x00, 0x00, 0x00, 0x00, // jne 0x7b6f0
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff // len=33
        }
    },

    // libc6_2.13-20ubuntu5.2_amd64.so
    {
        .data = {
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x28], rbp,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x20], r12,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x30], rbx,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x18], r13,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x10], r14,
            0x00, 0x00, // mov QWORD PTR [rsp-0x8], r15
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0xff, 0xff // len=33
        }
    },

    // libc6_2.17-93ubuntu2_amd64.so
    {
        .data = {
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x28], rbp,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x18], r13,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x30], rbx,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x20], r12,
            0x49, 0x89, 0xf5, // mov r13, rsi,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x10], r14,
            0x00, 0x00, // mov QWORD PTR [rsp-0x8], r15
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0xff, 0xff // len=33
        }
    },

    // musl-1.1.12-r8.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x8], rsi,
            0x00, 0x00, // jne 0x270a8,
            0x48, 0x89, 0xf7, // mov rdi, rsi,
            0xe8, 0xe3, 0x00, 0x00, 0x00, // call 0x26c86,
            0xe9, // jmp 0x271f1
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // libc6_2.19-13ubuntu3_amd64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x49, 0x89, 0xf6, // mov r14, rsi,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x08, // sub rsp, 0x8,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x33a58d] # 0x3bdfd8,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.31-17_amd64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x133645] # 0x1bdfe0,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6-amd64_2.6.1-1ubuntu9_i386.so
    {
        .data = {
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x18], r13,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x10], r14,
            0x49, 0x89, 0xfd, // mov r13, rdi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x30], rbx,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x28], rbp,
            0x49, 0x89, 0xf6, // mov r14, rsi,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x20], r12,
            0x00, 0x00, // mov QWORD PTR [rsp-0x8], r15
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0xff, 0xff // len=33
        }
    },

    // libc6-amd64_2.19-10ubuntu2_i386.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x49, 0x89, 0xf6, // mov r14, rsi,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x32685d] # 0x3a2fd8,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.24-7ubuntu2_amd64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x33ac55] # 0x3c0fd0,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // musl-1.2.2-3-omv4050.x86_64.so
    {
        .data = {
            0x50, // push rax,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x89, 0x00, 0x00, // mov QWORD PTR [rsp], rax,
            0xe8, 0x0b, 0x00, 0x00, 0x00, // call 0x604f1,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rcx, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x3b, 0x00, 0x00, // cmp rcx, QWORD PTR [rsp],
            0x00, // jne 0x609f7
        }, 
        .undef = {
            0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0xff // len=33
        }
    },

    // musl-1.1.22-r4.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x8], rsi,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x00, 0x00, // jne 0x23240,
            0x48, 0x89, 0xf7, // mov rdi, rsi,
            0xe8, 0xd8, 0x00, 0x00, 0x00, // call 0x15010,
            0x48, // mov rbx, rax
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // libc6_2.34-0ubuntu1_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x80, 0x3d, 0x09, 0x74, 0x17, 0x00, 0x00, // cmp BYTE PTR [rip+0x177409], 0x0 # 0x2201a8,
            0x00, 0x00, // je 0xa8ef8
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0xff, 0xff // len=33
        }
    },

    // libc6_2.3.2.ds1-13ubuntu2.2_amd64_2.so
    {
        .data = {
            0x48, 0x83, 0xec, 0x28, // sub rsp, 0x28,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rdx, QWORD PTR [rip+0x1c46dd] # 0x23b638,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x8], rbp,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x10], r12,
            0x48, 0x89, 0x00, 0x00, // mov QWORD PTR [rsp], rbx,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x18], r13,
            0x4c, 0x89, 0x00, // mov QWORD PTR [rsp+0x20], r14
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff // len=33
        }
    },

    // musl-1.1.24-r3.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x8], rsi,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x00, 0x00, // jne 0x2360d,
            0x48, 0x89, 0xf7, // mov rdi, rsi,
            0xe8, 0x0b, 0x00, 0x00, 0x00, // call 0x15010,
            0x49, // mov r12, rax
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // musl-1.2.2-r7.so
    {
        .data = {
            0xe9, 0x38, 0xfc, 0xff, 0xff, // jmp 0x24d43,
            0x48, 0x89, 0xf1, // mov rcx, rsi,
            0x48, 0x89, 0xd6, // mov rsi, rdx,
            0x48, 0x85, 0xd2, // test rdx, rdx,
            0x00, 0x00, // je 0x2511e,
            0x48, 0x89, 0xc8, // mov rax, rcx,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x70, 0x09, // jo 0x25127,
            0x48, 0x0f, 0xaf, 0xf1, // imul rsi, rcx,
            0xe9, 0x09, 0xff, 0xfe, 0xff, // jmp 0x15030
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // glibc-2.28-160.el8.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x338c77] # 0x3befd0,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax,
            0x00, 0x00, // jne 0x865a0
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff // len=33
        }
    },

    // libc6_2.23-0ubuntu5_amd64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf5, // mov r13, rsi,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x89, 0xfb, // mov rbx, rdi,
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x33f4a5] # 0x3c2fd0,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // musl_1.2.1-1_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x2ca50,
            0x48, 0xb8, 0xfe, // movabs rax, 0x7fffffffffffeffe
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.3.2.ds1-13ubuntu2.2_amd64.so
    {
        .data = {
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1c052d] # 0x234178,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x8], rbx,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x10], rbp,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x18], r12,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x20], r13,
            0x00, 0x00, // mov QWORD PTR [rsp+0x28], r14
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0xff, 0xff // len=33
        }
    },

    // libc-2.34-11.mga9.x86_64_3.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x49, 0x89, 0xf5, // mov r13, rsi,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x58, // sub rsp, 0x58,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x40d5] # 0x9fd0,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // musl-1.1.11-r5.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x8], rsi,
            0x00, 0x00, // jne 0x271ea,
            0x48, 0x89, 0xf7, // mov rdi, rsi,
            0xe8, 0xdd, 0x00, // call 0x26dbf
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff // len=33
        }
    },

    // libc6_2.21-0ubuntu4_amd64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x49, 0x89, 0xf6, // mov r14, rsi,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x33f59d] # 0x3c3fd8,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.17-317.el7.x86_64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x89, 0xfb, // mov rbx, rdi,
            0x48, 0x83, 0xec, 0x28, // sub rsp, 0x28,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x341405] # 0x3c6fd0,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.34-0-omv4050.x86_64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x80, 0x3d, 0x25, 0xdd, 0x16, 0x00, 0x00, // cmp BYTE PTR [rip+0x16dd25], 0x0 # 0x212ff0,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0xa53e0
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff // len=33
        }
    },

    // musl_1.1.24-1_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x8], rsi,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x2b6c0,
            0x48, // mov rbp, rdi
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // libc6-amd64_2.9-4ubuntu6.3_i386.so
    {
        .data = {
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x30], rbx,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x20], r12,
            0x48, 0x89, 0xfb, // mov rbx, rdi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x28], rbp,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x18], r13,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x4c, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp-0x10], r14,
            0x00, 0x00, // mov QWORD PTR [rsp-0x8], r15
        }, 
        .undef = {
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0xff, 0xff // len=33
        }
    },

    // musl-1.1.22-2-x86_64.so
    {
        .data = {
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x18a28,
            0x48, 0xb8, 0xdf, 0xef, 0xff, 0xff, 0xff, // movabs rax, 0x7fffffffffffefdf,
            0x00, 0x00, 0x00, // ,
            0x41, 0x57, // push r15,
            0x48, 0x8d, 0x56, 0xff, // lea rdx, [rsi-0x1],
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x48, // mov rbp, rdi
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00 // len=33
        }
    },

    // musl_0.9.15-1_amd64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x89, 0xfb, // mov rbx, rdi,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x24628,
            0x48, 0x8d, 0x56, 0xff, // lea rdx, [rsi-0x1],
            0x48, 0xb8, 0xdf, // movabs rax, 0x7fffffffffffefdf
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.34-0ubuntu2_amd64_2.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x49, 0x89, 0xf6, // mov r14, rsi,
            0x41, 0x55, // push r13,
            0x49, 0x89, 0xfd, // mov r13, rdi,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x78, // sub rsp, 0x78,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // 
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff // len=33
        }
    },

    // musl_1.1.5-2+deb8u1_amd64.so
    {
        .data = {
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x274e8,
            0x41, 0x57, // push r15,
            0x48, 0x8d, 0x56, 0xff, // lea rdx, [rsi-0x1],
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x48, 0xb8, 0xdf, 0xef, 0xff, 0xff, 0xff, // movabs rax, 0x7fffffffffffefdf,
            0x00, 0x00, 0x00, // ,
            0x55, // push rbp,
            0x53, // push rbx
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0xff, 0xff, 0xff,
            0x00,
            0x00 // len=33
        }
    },

    // musl-1.1.24-r10.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x8], rsi,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x00, 0x00, // jne 0x23662,
            0x48, 0x89, 0xf7, // mov rdi, rsi,
            0xe8, 0xb6, 0x00, 0x00, 0x00, // call 0x15010,
            0x49, // mov r12, rax
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // libc6_2.22-0ubuntu1_amd64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x49, 0x89, 0xf6, // mov r14, rsi,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x89, 0xfb, // mov rbx, rdi,
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x33fe0d] # 0x3c4fa8,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6-amd64_2.34-0ubuntu3_i386_2.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x49, 0x89, 0xf5, // mov r13, rsi,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xfc, // mov r12, rdi,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x68, // sub rsp, 0x68,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // 
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff // len=33
        }
    },

    // musl-1.1.19-r11.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x8], rsi,
            0x00, 0x00, // jne 0x2487b,
            0x48, 0x89, 0xf7, // mov rdi, rsi,
            0xe8, 0xd8, 0x00, 0x00, 0x00, // call 0x2444b,
            0x48, // mov rbx, rax
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // libc-2.34-7.fc35.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x80, 0x3d, 0x19, 0x42, 0x16, 0x00, 0x00, // cmp BYTE PTR [rip+0x164219], 0x0 # 0x201fe8,
            0x00, 0x00, // je 0x9df28
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0xff, 0xff // len=33
        }
    },

    // musl_1.1.21-2_amd64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x8], rsi,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x2a520,
            0x48, 0x89, 0xfb, // mov rbx, rdi,
            0x48, 0x8d, // lea rdi, [rsp+0x8]
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // libc-2.34-11.mga9.x86_64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x80, 0x3d, 0x0d, 0x75, 0x15, 0x00, 0x00, // cmp BYTE PTR [rip+0x15750d], 0x0 # 0x1ee1a8,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x96dc0
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff // len=33
        }
    },

    // libc6_2.34-0ubuntu3_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x80, 0x3d, 0xd9, 0x71, 0x17, 0x00, 0x00, // cmp BYTE PTR [rip+0x1771d9], 0x0 # 0x2201a8,
            0x00, 0x00, // je 0xa9128
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0xff, 0xff // len=33
        }
    },

    // musl-1.2.2-1-x86_64.so
    {
        .data = {
            0xe9, 0x50, 0x0a, 0x01, 0x00, // jmp 0x281a5,
            0x66, 0x2e, 0x0f, 0x1f, 0x84, 0x00, 0x00, // cs nop WORD PTR [rax+rax*1+0x0],
            0x00, 0x00, 0x00, // ,
            0x90, // nop,
            0x48, 0x89, 0xf1, // mov rcx, rsi,
            0x48, 0x89, 0xd6, // mov rsi, rdx,
            0x48, 0x85, 0xd2, // test rdx, rdx,
            0x00, 0x00, // je 0x17773,
            0x48, 0x89, 0xc8, // mov rax, rcx,
            0x48, 0xf7, 0xe6, // mul rsi
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0xff, 0xff, 0xff,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // musl-1.1.5-r5.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x89, 0xfb, // mov rbx, rdi,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x8], rsi,
            0x00, 0x00, // jne 0x26703,
            0x48, 0x89, 0xf7, // mov rdi, rsi,
            0xe8, 0xe9, 0x00, // call 0x262e4
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff // len=33
        }
    },

    // musl_1.1.19-1_amd64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x8], rsi,
            0x0f, 0x84, 0x00, 0x00, 0x00, 0x00, // je 0x2b258,
            0x48, 0x89, 0xfb, // mov rbx, rdi,
            0x48, 0x8d, // lea rdi, [rsp+0x8]
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00 // len=33
        }
    },

    // musl-1.2.1-2-x86_64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x49, 0x89, 0xf5, // mov r13, rsi,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x00, 0x00, // jne 0x27f33,
            0x48, 0x83, 0xc4, 0x18, // add rsp, 0x18,
            0x48, 0x89, 0xf7, // mov rdi, rsi,
            0x5b, // pop rbx,
            0x5d, // pop rbp,
            0x41, 0x5c, // pop r12
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00 // len=33
        }
    },

    // musl-1.1.20-r6.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x8], rsi,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x00, 0x00, // jne 0x24e5b,
            0x48, 0x89, 0xf7, // mov rdi, rsi,
            0xe8, 0xbd, 0x00, 0x00, 0x00, // call 0x17010,
            0x48, // mov rbx, rax
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // libc6-amd64_2.34-0ubuntu3_i386.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x80, 0x3d, 0x09, 0x43, 0x16, 0x00, 0x00, // cmp BYTE PTR [rip+0x164309], 0x0 # 0x2021a8,
            0x00, 0x00, // je 0x9dff8
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0xff, 0xff // len=33
        }
    },

    // musl-1.2.2-1-omv4002.x86_64.so
    {
        .data = {
            0x50, // push rax,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x89, 0x00, 0x00, // mov QWORD PTR [rsp], rax,
            0xe8, 0x25, 0x00, 0x00, 0x00, // call 0x6039e,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rcx, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x3b, 0x00, 0x00, // cmp rcx, QWORD PTR [rsp],
            0x00, // jne 0x6088a
        }, 
        .undef = {
            0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0xff // len=33
        }
    },

    // libc-2.34-0-omv4050.x86_64_2.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x49, 0x89, 0xf6, // mov r14, rsi,
            0x41, 0x55, // push r13,
            0x49, 0x89, 0xfd, // mov r13, rdi,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x78, // sub rsp, 0x78,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x89, 0x00, 0x00, // mov QWORD PTR [rsp+0x68], rax
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff // len=33
        }
    },

    // musl-1.1.15-r8.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x8], rsi,
            0x00, 0x00, // jne 0x241cc,
            0x48, 0x89, 0xf7, // mov rdi, rsi,
            0xe8, 0xea, 0x00, 0x00, 0x00, // call 0x23dae,
            0x48, // mov rbx, rax
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // libc6_2.23-0ubuntu11.2_amd64.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x89, 0xfb, // mov rbx, rdi,
            0x48, 0x83, 0xec, 0x38, // sub rsp, 0x38,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x33f8a5] # 0x3c3fd0,
            0x48, 0x8b, 0x00, // mov rax, QWORD PTR [rax],
            0x48, 0x85, 0xc0, // test rax, rax
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.34-1.2.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x49, 0x89, 0xf4, // mov r12, rsi,
            0x55, // push rbp,
            0x48, 0x89, 0xfd, // mov rbp, rdi,
            0x53, // push rbx,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x80, 0x3d, 0xe9, 0x42, 0x16, 0x00, 0x00, // cmp BYTE PTR [rip+0x1642e9], 0x0 # 0x201fe8,
            0x00, 0x00, // je 0x9de58
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0xff, 0xff // len=33
        }
    },

    // musl_1.2.2-4_amd64.so
    {
        .data = {
            0xe9, 0xfb, 0xfa, 0xff, 0xff, // jmp 0x29c80,
            0x66, 0x2e, 0x0f, 0x1f, 0x84, 0x00, 0x00, // cs nop WORD PTR [rax+rax*1+0x0],
            0x00, 0x00, 0x00, // ,
            0x90, // nop,
            0x48, 0x89, 0xf1, // mov rcx, rsi,
            0x48, 0x89, 0xd6, // mov rsi, rdx,
            0x48, 0x85, 0xd2, // test rdx, rdx,
            0x00, 0x00, // je 0x2a1a3,
            0x48, 0x89, 0xc8, // mov rax, rcx,
            0x48, 0xf7, 0xe6, // mul rsi
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0xff, 0xff, 0xff,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00 // len=33
        }
    },

    // musl-1.1.4-r5.so
    {
        .data = {
            0x41, 0x57, // push r15,
            0x41, 0x56, // push r14,
            0x41, 0x55, // push r13,
            0x41, 0x54, // push r12,
            0x55, // push rbp,
            0x53, // push rbx,
            0x48, 0x89, 0xfb, // mov rbx, rdi,
            0x48, 0x83, 0xec, 0x18, // sub rsp, 0x18,
            0x48, 0x85, 0xff, // test rdi, rdi,
            0x48, 0x89, 0x00, 0x00, 0x00, // mov QWORD PTR [rsp+0x8], rsi,
            0x00, 0x00, // jne 0x25cab,
            0x48, 0x89, 0xf7, // mov rdi, rsi,
            0xe8, 0xd3, 0x00, // call 0x25876
        }, 
        .undef = {
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00,
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff // len=33
        }
    },

    // musl-1.2.2-r3.so
    {
        .data = {
            0xe9, 0x38, 0xfc, 0xff, 0xff, // jmp 0x24d44,
            0x48, 0x89, 0xf1, // mov rcx, rsi,
            0x48, 0x89, 0xd6, // mov rsi, rdx,
            0x48, 0x85, 0xd2, // test rdx, rdx,
            0x00, 0x00, // je 0x2511f,
            0x48, 0x89, 0xc8, // mov rax, rcx,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x70, 0x09, // jo 0x25128,
            0x48, 0x0f, 0xaf, 0xf1, // imul rsi, rcx,
            0xe9, 0x08, 0xff, 0xfe, 0xff, // jmp 0x15030
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

};


static const uint FUNCSIGS_REALLOC_COUNT = 54;

static const funcsig FUNCSIGS_REALLOCARRAY[] = {

    // test-static2
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x188b4,
            0xe9, 0x0c, 0xed, 0xff, 0xff, // jmp 0x175c0,
            0x48, 0xc7, 0xc0, 0xc0, 0xff, 0xff, 0xff, // mov rax, 0xffffffffffffffc0,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.33-5.fc34.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8f954,
            0xe9, 0x2c, 0x6a, 0xf9, 0xff, // jmp 0x26380,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1334f5] # 0x1c2e50,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.32-0ubuntu3.2_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x9b9c4,
            0xe9, 0xbc, 0xa9, 0xf8, 0xff, // jmp 0x26380,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x14747d] # 0x1e2e48,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.31-17_amd64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8d4a0,
            0xe9, 0xa0, 0x7b, 0xf9, 0xff, // jmp 0x25040,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1309c9] # 0x1bde70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc-2.26-lp150.11.9.1.x86_64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x87960,
            0xe9, 0x70, 0x8b, 0xf9, 0xff, // jmp 0x204d0,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x32c511] # 0x3b3e78,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc-2.30-1-x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8e1e4,
            0xe9, 0xbc, 0x71, 0xf9, 0xff, // jmp 0x253a0,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x131c75] # 0x1bfe60,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.31-0ubuntu6_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0xa1ab4,
            0xe9, 0xbc, 0x38, 0xf8, 0xff, // jmp 0x25370,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1493b5] # 0x1eae70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.31-6.fc32.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8f6a4,
            0xe9, 0xcc, 0x5c, 0xf9, 0xff, // jmp 0x25370,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1337cd] # 0x1c2e78,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // musl-1.2.2-3-omv4050.x86_64.so
    {
        .data = {
            0x50, // push rax,
            0x48, 0x89, 0xd1, // mov rcx, rdx,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x89, 0x00, 0x00, // mov QWORD PTR [rsp], rax,
            0x48, 0x85, 0xd2, // test rdx, rdx,
            0x00, 0x00, // je 0x60a38,
            0x48, 0x89, 0xc8, // mov rax, rcx,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x71, 0x1e, // jno 0x60a38,
            0xe8, 0x8c, 0x00, // call 0x54cab
        }, 
        .undef = {
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0xff // len=33
        }
    },

    // libc6_2.32-0ubuntu6_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x9b7e4,
            0xe9, 0x9c, 0xab, 0xf8, 0xff, // jmp 0x26380,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x14665d] # 0x1e1e48,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.34-0ubuntu1_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0xaae24,
            0xe9, 0x6c, 0x15, 0xf8, 0xff, // jmp 0x2c390,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x16cfe5] # 0x217e10,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.29-0ubuntu2_amd64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x9c600,
            0xe9, 0x50, 0x8a, 0xf8, 0xff, // jmp 0x25050,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x147869] # 0x1e3e70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc6_2.27-5ubuntu1_amd64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x9e180,
            0xe9, 0xd0, 0x3e, 0xf8, 0xff, // jmp 0x22050,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x14ece1] # 0x1ece68,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc6_2.28-10_amd64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x874b0,
            0xe9, 0xa0, 0xab, 0xf9, 0xff, // jmp 0x22050,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1339b9] # 0x1bae70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // musl-1.2.2-r7.so
    {
        .data = {
            0x48, 0x89, 0xf1, // mov rcx, rsi,
            0x48, 0x89, 0xd6, // mov rsi, rdx,
            0x48, 0x85, 0xd2, // test rdx, rdx,
            0x00, 0x00, // je 0x2511e,
            0x48, 0x89, 0xc8, // mov rax, rcx,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x70, 0x09, // jo 0x25127,
            0x48, 0x0f, 0xaf, 0xf1, // imul rsi, rcx,
            0xe9, 0x09, 0xff, 0xfe, 0xff, // jmp 0x15030,
            0x50, // push rax,
            0xe8, 0x86, 0x00, 0x00, // call 0x1ceb3
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0xff, 0xff // len=33
        }
    },

    // libc-2.33-20.fc34.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8f6d4,
            0xe9, 0xac, 0x6c, 0xf9, 0xff, // jmp 0x26380,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x13377d] # 0x1c2e58,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.32-4-omv4002.x86_64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x98850,
            0xe9, 0xf0, 0xd7, 0xf8, 0xff, // jmp 0x26040,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x13e601] # 0x1d6e58,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // glibc-2.28-160.el8.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x88e24,
            0xe9, 0x6c, 0x8a, 0xf9, 0xff, // jmp 0x21890,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x33603d] # 0x3bee68,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.30-0ubuntu2_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0xa1b04,
            0xe9, 0x9c, 0x38, 0xf8, 0xff, // jmp 0x253a0,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x148365] # 0x1e9e70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.29-1-x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x06, // jo 0x87cf5,
            0xff, 0x25, 0x00, 0x00, 0x00, 0x00, // jmp QWORD PTR [rip+0x137063] # 0x1bed58,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x137144] # 0x1bee40,
            0x64, 0xc7, 0x00, 0x0c, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6-amd64_2.31-6_i386.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8d440,
            0xe9, 0x00, 0x7c, 0xf9, 0xff, // jmp 0x25040,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x130a29] # 0x1bde70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc-2.31-2.fc32.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8f944,
            0xe9, 0x2c, 0x5a, 0xf9, 0xff, // jmp 0x25370,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x13352d] # 0x1c2e78,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.34-0-omv4050.x86_64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0xa6e50,
            0xe9, 0xe0, 0x51, 0xf8, 0xff, // jmp 0x2c030,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x163fa1] # 0x20adf8,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc-2.29-20.mga7.x86_64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8d280,
            0xe9, 0xd0, 0x7d, 0xf9, 0xff, // jmp 0x25050,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rdx, QWORD PTR fs:0x0,
            0x00, 0x00, // ,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x135be0] # 0x1c2e70,
            0xc7, // mov DWORD PTR [rdx+rax*1], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // libc-2.29-3-x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8a964,
            0xe9, 0x1c, 0xaa, 0xf9, 0xff, // jmp 0x25380,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x131505] # 0x1bbe70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.27-3ubuntu1.2_amd64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x9d110,
            0xe9, 0x10, 0x3f, 0xf8, 0xff, // jmp 0x21020,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x34dd51] # 0x3eae68,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc6_2.33-0ubuntu5_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x9b684,
            0xe9, 0xfc, 0xac, 0xf8, 0xff, // jmp 0x26380,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1447bd] # 0x1dfe48,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6-amd64_2.33-0ubuntu5_i386.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8f884,
            0xe9, 0xfc, 0x6a, 0xf9, 0xff, // jmp 0x26380,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1335bd] # 0x1c2e48,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.26-lp151.19.15.1.x86_64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8ae90,
            0xe9, 0x40, 0x8a, 0xf9, 0xff, // jmp 0x238d0,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x328fe1] # 0x3b3e78,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc6-amd64_2.31-0ubuntu9_i386.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8f514,
            0xe9, 0x5c, 0x5e, 0xf9, 0xff, // jmp 0x25370,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x132955] # 0x1c1e70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.26-0ubuntu2_amd64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x93350,
            0xe9, 0x30, 0xd2, 0xf8, 0xff, // jmp 0x20580,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x346b11] # 0x3d9e68,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc6_2.31-0ubuntu10_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0xa1b34,
            0xe9, 0x3c, 0x38, 0xf8, 0xff, // jmp 0x25370,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x149335] # 0x1eae70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.26-lp152.26.6.1.x86_64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8adb0,
            0xe9, 0x20, 0x8b, 0xf9, 0xff, // jmp 0x238d0,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x3290c1] # 0x3b3e78,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc-2.31-3-x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8e7c4,
            0xe9, 0xac, 0x6b, 0xf9, 0xff, // jmp 0x25370,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1316a5] # 0x1bfe70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.32-18.mga8.x86_64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x895e0,
            0xe9, 0x60, 0xca, 0xf9, 0xff, // jmp 0x26040,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rdx, QWORD PTR fs:0x0,
            0x00, 0x00, // ,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x129858] # 0x1b2e48,
            0xc7, // mov DWORD PTR [rdx+rax*1], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // libc6-amd64_2.26-0ubuntu2.1_i386.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x84ce0,
            0xe9, 0xf0, 0xb7, 0xf9, 0xff, // jmp 0x204d0,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x32a181] # 0x3aee68,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc6_2.31-2ubuntu1_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x9bc74,
            0xe9, 0xfc, 0x96, 0xf8, 0xff, // jmp 0x25370,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1451f5] # 0x1e0e70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.30-0ubuntu3_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0xa1b34,
            0xe9, 0x6c, 0x38, 0xf8, 0xff, // jmp 0x253a0,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x148335] # 0x1e9e70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.32-4-x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8f124,
            0xe9, 0x5c, 0x72, 0xf9, 0xff, // jmp 0x26380,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x132d25] # 0x1c1e50,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.31-1-x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8e1c4,
            0xe9, 0xac, 0x71, 0xf9, 0xff, // jmp 0x25370,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x130ca5] # 0x1bee70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.27-3ubuntu1_amd64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x9d0a0,
            0xe9, 0x80, 0x3f, 0xf8, 0xff, // jmp 0x21020,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x34ddc1] # 0x3eae68,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc6_2.33-0experimental2_amd64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8d7e0,
            0xe9, 0x60, 0x88, 0xf9, 0xff, // jmp 0x26040,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x12f669] # 0x1bce50,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc6_2.33-0ubuntu2_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x9b634,
            0xe9, 0x4c, 0xad, 0xf8, 0xff, // jmp 0x26380,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x14480d] # 0x1dfe48,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6-amd64_2.30-0ubuntu2_i386.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8f594,
            0xe9, 0x0c, 0x5e, 0xf9, 0xff, // jmp 0x253a0,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1338d5] # 0x1c2e70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.32-10.fc33.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8fbc4,
            0xe9, 0xbc, 0x67, 0xf9, 0xff, // jmp 0x26380,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x13428d] # 0x1c3e58,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.34-7.fc35.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x9f454,
            0xe9, 0x3c, 0xcf, 0xf8, 0xff, // jmp 0x2c390,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x15a9ad] # 0x1f9e08,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.32-15.mga8.x86_64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x89610,
            0xe9, 0x30, 0xca, 0xf9, 0xff, // jmp 0x26040,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rdx, QWORD PTR fs:0x0,
            0x00, 0x00, // ,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x129828] # 0x1b2e48,
            0xc7, // mov DWORD PTR [rdx+rax*1], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // libc-2.34-11.mga9.x86_64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x980c0,
            0xe9, 0x70, 0x3f, 0xf9, 0xff, // jmp 0x2c030,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rdx, QWORD PTR fs:0x0,
            0x00, 0x00, // ,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x14dd40] # 0x1e5e10,
            0xc7, // mov DWORD PTR [rdx+rax*1], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // libc6_2.34-0ubuntu3_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0xab054,
            0xe9, 0x3c, 0x13, 0xf8, 0xff, // jmp 0x2c390,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x16cdb5] # 0x217e10,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.31-4-x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8e834,
            0xe9, 0x3c, 0x6b, 0xf9, 0xff, // jmp 0x25370,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x131635] # 0x1bfe70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.31-7.30.x86_64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8d2f0,
            0xe9, 0x30, 0x76, 0xf9, 0xff, // jmp 0x24920,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x340b79] # 0x3cde70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // musl-1.2.2-1-x86_64.so
    {
        .data = {
            0x48, 0x89, 0xf1, // mov rcx, rsi,
            0x48, 0x89, 0xd6, // mov rsi, rdx,
            0x48, 0x85, 0xd2, // test rdx, rdx,
            0x00, 0x00, // je 0x17773,
            0x48, 0x89, 0xc8, // mov rax, rcx,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x70, 0x09, // jo 0x1777c,
            0x48, 0x0f, 0xaf, 0xf1, // imul rsi, rcx,
            0xe9, 0xb4, 0xd8, 0xff, 0xff, // jmp 0x15030,
            0x50, // push rax,
            0xe8, 0xb3, 0x00, 0x00, // call 0x21e35
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0xff, 0xff // len=33
        }
    },

    // libc-2.29-23.mga7.x86_64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8d1f0,
            0xe9, 0x60, 0x7e, 0xf9, 0xff, // jmp 0x25050,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rdx, QWORD PTR fs:0x0,
            0x00, 0x00, // ,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x135c70] # 0x1c2e70,
            0xc7, // mov DWORD PTR [rdx+rax*1], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // libc-2.26-lp150.11.20.1.x86_64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x879e0,
            0xe9, 0x30, 0x8b, 0xf9, 0xff, // jmp 0x20510,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x32c491] # 0x3b3e78,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc6-amd64_2.32-0ubuntu3_i386.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8fd54,
            0xe9, 0x2c, 0x66, 0xf9, 0xff, // jmp 0x26380,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1350ed] # 0x1c4e48,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.27-3ubuntu1.3_amd64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x9d1e0,
            0xe9, 0x40, 0x3e, 0xf8, 0xff, // jmp 0x21020,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x34dc81] # 0x3eae68,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc-2.33-5-x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8e894,
            0xe9, 0xec, 0x7a, 0xf9, 0xff, // jmp 0x26380,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1315b5] # 0x1bfe50,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.32-0ubuntu4_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x9b7a4,
            0xe9, 0xdc, 0xab, 0xf8, 0xff, // jmp 0x26380,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x14669d] # 0x1e1e48,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6-amd64_2.27-3ubuntu1_i386.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x85600,
            0xe9, 0x20, 0xba, 0xf9, 0xff, // jmp 0x21020,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x32e861] # 0x3b3e68,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc-2.33-3-x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8eb64,
            0xe9, 0x1c, 0x78, 0xf9, 0xff, // jmp 0x26380,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1322e5] # 0x1c0e50,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6-amd64_2.32-4_i386.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8dad0,
            0xe9, 0x70, 0x85, 0xf9, 0xff, // jmp 0x26040,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x130371] # 0x1bde48,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc6-amd64_2.27-3ubuntu1.2_i386.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x85640,
            0xe9, 0xe0, 0xb9, 0xf9, 0xff, // jmp 0x21020,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x32e821] # 0x3b3e68,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc6-amd64_2.28-10_i386.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x874a0,
            0xe9, 0xb0, 0xab, 0xf9, 0xff, // jmp 0x22050,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1339c9] # 0x1bae70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc-2.33-4-x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8eb24,
            0xe9, 0x5c, 0x78, 0xf9, 0xff, // jmp 0x26380,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x132325] # 0x1c0e50,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6_2.26-0ubuntu2.1_amd64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x933b0,
            0xe9, 0xd0, 0xd1, 0xf8, 0xff, // jmp 0x20580,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x346ab1] # 0x3d9e68,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc-2.32-1.fc33.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8fd14,
            0xe9, 0x6c, 0x66, 0xf9, 0xff, // jmp 0x26380,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x13513d] # 0x1c4e58,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc6-amd64_2.27-3ubuntu1.4_i386.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x85670,
            0xe9, 0xb0, 0xb9, 0xf9, 0xff, // jmp 0x21020,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x32e7f1] # 0x3b3e68,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc6-amd64_2.28-0ubuntu1_i386.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x87760,
            0xe9, 0xf0, 0xa8, 0xf9, 0xff, // jmp 0x22050,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x133709] # 0x1bae70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc-2.29-21.mga7.x86_64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8d1c0,
            0xe9, 0x90, 0x7e, 0xf9, 0xff, // jmp 0x25050,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rdx, QWORD PTR fs:0x0,
            0x00, 0x00, // ,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x135ca0] # 0x1c2e70,
            0xc7, // mov DWORD PTR [rdx+rax*1], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00 // len=33
        }
    },

    // libc6-amd64_2.34-0ubuntu3_i386.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x9f4e4,
            0xe9, 0xac, 0xce, 0xf8, 0xff, // jmp 0x2c390,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x15a925] # 0x1f9e10,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // musl-1.2.2-1-omv4002.x86_64.so
    {
        .data = {
            0x50, // push rax,
            0x48, 0x89, 0xd1, // mov rcx, rdx,
            0x64, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR fs:0x28,
            0x00, 0x00, // ,
            0x48, 0x89, 0x00, 0x00, // mov QWORD PTR [rsp], rax,
            0x48, 0x85, 0xd2, // test rdx, rdx,
            0x00, 0x00, // je 0x608cb,
            0x48, 0x89, 0xc8, // mov rax, rcx,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x71, 0x1e, // jno 0x608cb,
            0xe8, 0xa6, 0x00, // call 0x54c58
        }, 
        .undef = {
            0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0xff, 0xff,
            0x00, 0x00, 0xff, 0xff,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0xff // len=33
        }
    },

    // libc6_2.31-3_amd64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8d560,
            0xe9, 0xe0, 0x7a, 0xf9, 0xff, // jmp 0x25040,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x130909] # 0x1bde70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc6_2.26-0ubuntu4_amd64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x95bf0,
            0xe9, 0xe0, 0xa8, 0xf8, 0xff, // jmp 0x204d0,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x348271] # 0x3dde68,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc-2.34-1.2.x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x9f384,
            0xe9, 0x0c, 0xd0, 0xf8, 0xff, // jmp 0x2c390,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x15aa6d] # 0x1f9df8,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // musl_1.2.2-4_amd64.so
    {
        .data = {
            0x48, 0x89, 0xf1, // mov rcx, rsi,
            0x48, 0x89, 0xd6, // mov rsi, rdx,
            0x48, 0x85, 0xd2, // test rdx, rdx,
            0x00, 0x00, // je 0x2a1a3,
            0x48, 0x89, 0xc8, // mov rax, rcx,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x70, 0x09, // jo 0x2a1ac,
            0x48, 0x0f, 0xaf, 0xf1, // imul rsi, rcx,
            0xe9, 0x84, 0xae, 0xfe, 0xff, // jmp 0x15030,
            0x50, // push rax,
            0xe8, 0x4e, 0x00, 0x00, // call 0x1b400
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0xff, 0xff // len=33
        }
    },

    // libc6_2.28-0ubuntu1_amd64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x99740,
            0xe9, 0x10, 0x89, 0xf8, 0xff, // jmp 0x22050,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x14a729] # 0x1e3e70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc6_2.26-0ubuntu1_amd64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x93360,
            0xe9, 0x20, 0xd2, 0xf8, 0xff, // jmp 0x20580,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x346b01] # 0x3d9e68,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // musl-1.2.2-r3.so
    {
        .data = {
            0x48, 0x89, 0xf1, // mov rcx, rsi,
            0x48, 0x89, 0xd6, // mov rsi, rdx,
            0x48, 0x85, 0xd2, // test rdx, rdx,
            0x00, 0x00, // je 0x2511f,
            0x48, 0x89, 0xc8, // mov rax, rcx,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x70, 0x09, // jo 0x25128,
            0x48, 0x0f, 0xaf, 0xf1, // imul rsi, rcx,
            0xe9, 0x08, 0xff, 0xfe, 0xff, // jmp 0x15030,
            0x50, // push rax,
            0xe8, 0x85, 0x00, 0x00, // call 0x1ceb3
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0xff, 0xff // len=33
        }
    },

    // libc6_2.29-0ubuntu1_amd64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x9c670,
            0xe9, 0xe0, 0x89, 0xf8, 0xff, // jmp 0x25050,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1477f9] # 0x1e3e70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // musl_1.2.2-1_amd64.so
    {
        .data = {
            0x48, 0x89, 0xf1, // mov rcx, rsi,
            0x48, 0x89, 0xd6, // mov rsi, rdx,
            0x48, 0x85, 0xd2, // test rdx, rdx,
            0x00, 0x00, // je 0x2c743,
            0x48, 0x89, 0xc8, // mov rax, rcx,
            0x48, 0xf7, 0xe6, // mul rsi,
            0x70, 0x09, // jo 0x2c74c,
            0x48, 0x0f, 0xaf, 0xf1, // imul rsi, rcx,
            0xe9, 0xe4, 0x88, 0xfe, 0xff, // jmp 0x15030,
            0x50, // push rax,
            0xe8, 0x5e, 0x00, 0x00, // call 0x1f1b0
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0xff, 0xff,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00,
            0x00, 0x00, 0xff, 0xff // len=33
        }
    },

    // libc6-amd64_2.26-0ubuntu2_i386.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x84cc0,
            0xe9, 0x10, 0xb8, 0xf9, 0xff, // jmp 0x204d0,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x32a1a1] # 0x3aee68,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc-2.29-2-x86_64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x8a7c4,
            0xe9, 0xbc, 0xab, 0xf9, 0xff, // jmp 0x25380,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x1316a5] # 0x1bbe70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

    // libc-2.33-2-omv4002.x86_64.so
    {
        .data = {
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x988d0,
            0xe9, 0x70, 0xd7, 0xf8, 0xff, // jmp 0x26040,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x13d579] # 0x1d5e50,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc,
            0x31, 0xc0, // xor eax, eax,
            0xc3, // ret
        }, 
        .undef = {
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00 // len=33
        }
    },

    // libc6_2.29-0ubuntu3_amd64.so
    {
        .data = {
            0xf3, 0x0f, 0x1e, 0xfa, // endbr64,
            0x48, 0x89, 0xf0, // mov rax, rsi,
            0x48, 0xf7, 0xe2, // mul rdx,
            0x48, 0x89, 0xc6, // mov rsi, rax,
            0x70, 0x05, // jo 0x9ec64,
            0xe9, 0x1c, 0x67, 0xf8, 0xff, // jmp 0x25380,
            0x48, 0x8b, 0x00, 0x00, 0x00, 0x00, 0x00, // mov rax, QWORD PTR [rip+0x14a205] # 0x1e8e70,
            0x64, 0xc7, 0x00, 0x0c, 0x00, 0x00, // mov DWORD PTR fs:[rax], 0xc
        }, 
        .undef = {
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00 // len=33
        }
    },

};


static const uint FUNCSIGS_REALLOCARRAY_COUNT = 84;

#endif
